# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from monai.transforms.traits import InvertibleTrait

__all__ = ["ApplyPending"]


class ApplyPending(InvertibleTrait):
    """
    ApplyPending can be inserted into a pipeline that is being executed lazily in order to ensure
    resampling happens before the next transform. It doesn't do anything itself, but its presence
    causes the pipeline to be executed as ApplyPending doesn't implement ```LazyTrait``.

    See ``Compose`` for a detailed explanation of the lazy resampling feature.
    """

    def __call__(self, data):
        return data

    def inverse(self, data):
        return data



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import unittest

import torch
from parameterized import parameterized

from monai.networks import eval_mode
from monai.networks.nets import VNet
from tests.test_utils import test_script_save

device = "cuda" if torch.cuda.is_available() else "cpu"

TEST_CASE_VNET_2D_1 = [
    {"spatial_dims": 2, "in_channels": 4, "out_channels": 1, "act": "elu", "dropout_dim": 1},
    (1, 4, 32, 32),
    (1, 1, 32, 32),
]
TEST_CASE_VNET_2D_2 = [
    {"spatial_dims": 2, "in_channels": 2, "out_channels": 2, "act": "prelu", "dropout_dim": 2},
    (1, 2, 32, 32),
    (1, 2, 32, 32),
]
TEST_CASE_VNET_2D_3 = [
    {"spatial_dims": 2, "in_channels": 1, "out_channels": 3, "dropout_dim": 3},
    (1, 1, 32, 32),
    (1, 3, 32, 32),
]
TEST_CASE_VNET_3D_1 = [
    {"spatial_dims": 3, "in_channels": 4, "out_channels": 1, "act": "elu", "dropout_dim": 1},
    (1, 4, 32, 32, 32),
    (1, 1, 32, 32, 32),
]
TEST_CASE_VNET_3D_2 = [
    {"spatial_dims": 3, "in_channels": 2, "out_channels": 2, "act": "prelu", "dropout_dim": 2},
    (1, 2, 32, 32, 32),
    (1, 2, 32, 32, 32),
]
TEST_CASE_VNET_3D_3 = [
    {"spatial_dims": 3, "in_channels": 1, "out_channels": 3, "dropout_dim": 3},
    (1, 1, 32, 32, 32),
    (1, 3, 32, 32, 32),
]


class TestVNet(unittest.TestCase):
    @parameterized.expand(
        [
            TEST_CASE_VNET_2D_1,
            TEST_CASE_VNET_2D_2,
            TEST_CASE_VNET_2D_3,
            TEST_CASE_VNET_3D_1,
            TEST_CASE_VNET_3D_2,
            TEST_CASE_VNET_3D_3,
        ]
    )
    def test_vnet_shape(self, input_param, input_shape, expected_shape):
        net = VNet(**input_param).to(device)
        with eval_mode(net):
            result = net.forward(torch.randn(input_shape).to(device))
            self.assertEqual(result.shape, expected_shape)

    def test_script(self):
        net = VNet(spatial_dims=3, in_channels=1, out_channels=3, dropout_dim=3)
        test_data = torch.randn(1, 1, 32, 32, 32)
        test_script_save(net, test_data)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import unittest

import torch
from parameterized import parameterized

from monai.networks.schedulers import PNDMScheduler
from tests.test_utils import assert_allclose

TEST_2D_CASE = []
for beta_schedule in ["linear_beta", "scaled_linear_beta"]:
    TEST_2D_CASE.append([{"schedule": beta_schedule}, (2, 6, 16, 16), (2, 6, 16, 16)])

TEST_3D_CASE = []
for beta_schedule in ["linear_beta", "scaled_linear_beta"]:
    TEST_3D_CASE.append([{"schedule": beta_schedule}, (2, 6, 16, 16, 16), (2, 6, 16, 16, 16)])

TEST_CASES = TEST_2D_CASE + TEST_3D_CASE

TEST_FULl_LOOP = [
    [
        {"schedule": "linear_beta"},
        (1, 1, 2, 2),
        torch.Tensor([[[[-2123055.2500, -459014.2812], [2863438.0000, -1263401.7500]]]]),
    ]
]


class TestDDPMScheduler(unittest.TestCase):
    @parameterized.expand(TEST_CASES)
    def test_add_noise(self, input_param, input_shape, expected_shape):
        scheduler = PNDMScheduler(**input_param)
        original_sample = torch.zeros(input_shape)
        noise = torch.randn_like(original_sample)
        timesteps = torch.randint(0, scheduler.num_train_timesteps, (original_sample.shape[0],)).long()
        noisy = scheduler.add_noise(original_samples=original_sample, noise=noise, timesteps=timesteps)
        self.assertEqual(noisy.shape, expected_shape)

    @parameterized.expand(TEST_CASES)
    def test_step_shape(self, input_param, input_shape, expected_shape):
        scheduler = PNDMScheduler(**input_param)
        scheduler.set_timesteps(600)
        model_output = torch.randn(input_shape)
        sample = torch.randn(input_shape)
        output_step = scheduler.step(model_output=model_output, timestep=500, sample=sample)
        self.assertEqual(output_step[0].shape, expected_shape)
        self.assertEqual(output_step[1], None)

    @parameterized.expand(TEST_FULl_LOOP)
    def test_full_timestep_loop(self, input_param, input_shape, expected_output):
        scheduler = PNDMScheduler(**input_param)
        scheduler.set_timesteps(50)
        torch.manual_seed(42)
        model_output = torch.randn(input_shape)
        sample = torch.randn(input_shape)
        for t in range(50):
            sample, _ = scheduler.step(model_output=model_output, timestep=t, sample=sample)
        assert_allclose(sample, expected_output, rtol=1e-3, atol=1e-3)

    @parameterized.expand(TEST_FULl_LOOP)
    def test_timestep_two_loops(self, input_param, input_shape, expected_output):
        scheduler = PNDMScheduler(**input_param)
        scheduler.set_timesteps(50)
        torch.manual_seed(42)
        model_output = torch.randn(input_shape)
        sample = torch.randn(input_shape)
        for t in range(50):
            sample, _ = scheduler.step(model_output=model_output, timestep=t, sample=sample)
        torch.manual_seed(42)
        model_output2 = torch.randn(input_shape)
        sample2 = torch.randn(input_shape)
        scheduler.set_timesteps(50)
        for t in range(50):
            sample2, _ = scheduler.step(model_output=model_output2, timestep=t, sample=sample2)
        assert_allclose(sample, sample2, rtol=1e-3, atol=1e-3)

    def test_set_timesteps(self):
        scheduler = PNDMScheduler(num_train_timesteps=1000, skip_prk_steps=True)
        scheduler.set_timesteps(num_inference_steps=100)
        self.assertEqual(scheduler.num_inference_steps, 100)
        self.assertEqual(len(scheduler.timesteps), 100)

    def test_set_timesteps_prk(self):
        scheduler = PNDMScheduler(num_train_timesteps=1000, skip_prk_steps=False)
        scheduler.set_timesteps(num_inference_steps=100)
        self.assertEqual(scheduler.num_inference_steps, 109)
        self.assertEqual(len(scheduler.timesteps), 109)

    def test_set_timesteps_with_num_inference_steps_bigger_than_num_train_timesteps(self):
        scheduler = PNDMScheduler(num_train_timesteps=1000)
        with self.assertRaises(ValueError):
            scheduler.set_timesteps(num_inference_steps=2000)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import unittest

import numpy as np
import torch
from parameterized import parameterized

from monai.losses.deform import BendingEnergyLoss

device = "cuda" if torch.cuda.is_available() else "cpu"

TEST_CASES = [
    [{}, {"pred": torch.ones((1, 3, 5, 5, 5), device=device)}, 0.0],
    [{}, {"pred": torch.arange(0, 5, device=device)[None, None, None, None, :].expand(1, 3, 5, 5, 5)}, 0.0],
    [
        {"normalize": False},
        {"pred": torch.arange(0, 5, device=device)[None, None, None, None, :].expand(1, 3, 5, 5, 5) ** 2},
        4.0,
    ],
    [
        {"normalize": False},
        {"pred": torch.arange(0, 5, device=device)[None, None, None, :].expand(1, 2, 5, 5) ** 2},
        4.0,
    ],
    [{"normalize": False}, {"pred": torch.arange(0, 5, device=device)[None, None, :].expand(1, 1, 5) ** 2}, 4.0],
    [
        {"normalize": True},
        {"pred": torch.arange(0, 5, device=device)[None, None, None, None, :].expand(1, 3, 5, 5, 5) ** 2},
        100.0,
    ],
    [
        {"normalize": True},
        {"pred": torch.arange(0, 5, device=device)[None, None, None, :].expand(1, 2, 5, 5) ** 2},
        100.0,
    ],
    [{"normalize": True}, {"pred": torch.arange(0, 5, device=device)[None, None, :].expand(1, 1, 5) ** 2}, 100.0],
]


class TestBendingEnergy(unittest.TestCase):

    @parameterized.expand(TEST_CASES)
    def test_shape(self, input_param, input_data, expected_val):
        result = BendingEnergyLoss(**input_param).forward(**input_data)
        np.testing.assert_allclose(result.detach().cpu().numpy(), expected_val, rtol=1e-5)

    def test_ill_shape(self):
        loss = BendingEnergyLoss()
        # not in 3-d, 4-d, 5-d
        with self.assertRaisesRegex(ValueError, "Expecting 3-d, 4-d or 5-d"):
            loss.forward(torch.ones((1, 3), device=device))
        with self.assertRaisesRegex(ValueError, "Expecting 3-d, 4-d or 5-d"):
            loss.forward(torch.ones((1, 4, 5, 5, 5, 5), device=device))
        with self.assertRaisesRegex(ValueError, "All spatial dimensions"):
            loss.forward(torch.ones((1, 3, 4, 5, 5), device=device))
        with self.assertRaisesRegex(ValueError, "All spatial dimensions"):
            loss.forward(torch.ones((1, 3, 5, 4, 5)))
        with self.assertRaisesRegex(ValueError, "All spatial dimensions"):
            loss.forward(torch.ones((1, 3, 5, 5, 4)))

        # number of vector components unequal to number of spatial dims
        with self.assertRaisesRegex(ValueError, "Number of vector components"):
            loss.forward(torch.ones((1, 2, 5, 5, 5)))
        with self.assertRaisesRegex(ValueError, "Number of vector components"):
            loss.forward(torch.ones((1, 2, 5, 5, 5)))

    def test_ill_opts(self):
        pred = torch.rand(1, 3, 5, 5, 5).to(device=device)
        with self.assertRaisesRegex(ValueError, ""):
            BendingEnergyLoss(reduction="unknown")(pred)
        with self.assertRaisesRegex(ValueError, ""):
            BendingEnergyLoss(reduction=None)(pred)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import tempfile
import unittest

import torch
from parameterized import parameterized

from monai.handlers import TrtHandler
from monai.networks import trt_compile
from monai.networks.nets import cell_sam_wrapper, vista3d132
from monai.utils import min_version, optional_import
from tests.test_utils import SkipIfBeforeComputeCapabilityVersion, skip_if_no_cuda, skip_if_quick, skip_if_windows

trt, trt_imported = optional_import("tensorrt", "10.1.0", min_version)
torch_tensorrt, torch_trt_imported = optional_import("torch_tensorrt")
polygraphy, polygraphy_imported = optional_import("polygraphy")
build_sam_vit_b, has_sam = optional_import("segment_anything.build_sam", name="build_sam_vit_b")

TEST_CASE_1 = ["fp32"]
TEST_CASE_2 = ["fp16"]


class ListAdd(torch.nn.Module):
    def __init__(self):
        super().__init__()

    def forward(self, x: list[torch.Tensor], y: torch.Tensor, z: torch.Tensor, bs: float = 0.1):
        y1 = y.clone()
        x1 = x.copy()
        z1 = z + y
        for xi in x:
            y1 = y1 + xi + bs
        return x1, [y1, z1], y1 + z1


@skip_if_windows
@skip_if_no_cuda
@skip_if_quick
@unittest.skipUnless(trt_imported, "tensorrt is required")
@unittest.skipUnless(polygraphy_imported, "polygraphy is required")
@SkipIfBeforeComputeCapabilityVersion((7, 5))
class TestTRTCompile(unittest.TestCase):
    def setUp(self):
        self.gpu_device = torch.cuda.current_device()

    def tearDown(self):
        current_device = torch.cuda.current_device()
        if current_device != self.gpu_device:
            torch.cuda.set_device(self.gpu_device)

    # @unittest.skipUnless(torch_trt_imported, "torch_tensorrt is required")
    def test_handler(self):
        from ignite.engine import Engine

        net1 = torch.nn.Sequential(*[torch.nn.PReLU(), torch.nn.PReLU()])
        data1 = net1.state_dict()
        data1["0.weight"] = torch.tensor([0.1])
        data1["1.weight"] = torch.tensor([0.2])
        net1.load_state_dict(data1)
        net1.cuda()

        with tempfile.TemporaryDirectory() as tempdir:
            engine = Engine(lambda e, b: None)
            args = {"method": "onnx", "dynamic_batchsize": [1, 4, 8]}
            TrtHandler(net1, tempdir + "/trt_handler", args=args).attach(engine)
            engine.run([0] * 8, max_epochs=1)
            self.assertIsNotNone(net1._trt_compiler)
            self.assertIsNone(net1._trt_compiler.engine)
            net1.forward(torch.tensor([[0.0, 1.0], [1.0, 2.0]], device="cuda"))
            self.assertIsNotNone(net1._trt_compiler.engine)

    def test_lists(self):
        model = ListAdd().cuda()

        with torch.no_grad(), tempfile.TemporaryDirectory() as tmpdir:
            args = {
                "output_lists": [[-1], [2], []],
                "export_args": {"dynamo": False, "verbose": True},
                "dynamic_batchsize": [1, 4, 8],
            }
            x = torch.randn(1, 16).to("cuda")
            y = torch.randn(1, 16).to("cuda")
            z = torch.randn(1, 16).to("cuda")
            input_example = ([x, y, z], y.clone(), z.clone())
            output_example = model(*input_example)
            trt_compile(model, f"{tmpdir}/test_lists", args=args)
            self.assertIsNone(model._trt_compiler.engine)
            trt_output = model(*input_example)
            # Check that lazy TRT build succeeded
            self.assertIsNotNone(model._trt_compiler.engine)
            torch.testing.assert_close(trt_output, output_example, rtol=0.01, atol=0.01)

    @parameterized.expand([TEST_CASE_1, TEST_CASE_2])
    @unittest.skipUnless(has_sam, "Requires SAM installation")
    def test_cell_sam_wrapper_value(self, precision):
        model = cell_sam_wrapper.CellSamWrapper(checkpoint=None).to("cuda")
        with torch.no_grad(), tempfile.TemporaryDirectory() as tmpdir:
            model.eval()
            input_example = torch.randn(1, 3, 128, 128).to("cuda")
            output_example = model(input_example)
            trt_compile(model, f"{tmpdir}/test_cell_sam_wrapper_trt_compile", args={"precision": precision})
            self.assertIsNone(model._trt_compiler.engine)
            trt_output = model(input_example)
            # Check that lazy TRT build succeeded
            self.assertIsNotNone(model._trt_compiler.engine)
            torch.testing.assert_close(trt_output, output_example, rtol=0.01, atol=0.01)

    @parameterized.expand([TEST_CASE_1, TEST_CASE_2])
    def test_vista3d(self, precision):
        model = vista3d132(in_channels=1).to("cuda")
        with torch.no_grad(), tempfile.TemporaryDirectory() as tmpdir:
            model.eval()
            input_example = torch.randn(1, 1, 64, 64, 64).to("cuda")
            output_example = model(input_example)
            model = trt_compile(
                model,
                f"{tmpdir}/test_vista3d_trt_compile",
                args={"precision": precision, "dynamic_batchsize": [1, 2, 4]},
                submodule=["image_encoder.encoder", "class_head"],
            )
            self.assertIsNotNone(model.image_encoder.encoder._trt_compiler)
            self.assertIsNotNone(model.class_head._trt_compiler)
            trt_output = model.forward(input_example)
            # Check that lazy TRT build succeeded
            # TODO: set up input_example in such a way that image_encoder.encoder and class_head are called
            # and uncomment the asserts below
            # self.assertIsNotNone(model.image_encoder.encoder._trt_compiler.engine)
            # self.assertIsNotNone(model.class_head._trt_compiler.engine)
            torch.testing.assert_close(trt_output, output_example, rtol=0.01, atol=0.01)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import sys
import unittest
from copy import deepcopy

import numpy as np
import torch

from monai.data import DataLoader, Dataset, MetaTensor, create_test_image_3d, decollate_batch
from monai.transforms import (
    CastToType,
    Compose,
    EnsureChannelFirst,
    Invert,
    Lambda,
    LoadImage,
    Orientation,
    RandAffine,
    RandAxisFlip,
    RandFlip,
    RandRotate,
    RandRotate90,
    RandZoom,
    ResizeWithPadOrCrop,
    Spacing,
)
from monai.utils import set_determinism
from tests.test_utils import assert_allclose, make_nifti_image


class TestInvert(unittest.TestCase):
    def test_invert(self):
        set_determinism(seed=0)
        im_fname = make_nifti_image(create_test_image_3d(101, 100, 107, noise_max=100)[1])  # label image, discrete
        data = [im_fname for _ in range(12)]
        transform = Compose(
            [
                LoadImage(image_only=True),
                EnsureChannelFirst(),
                Orientation("RPS"),
                Spacing(pixdim=(1.2, 1.01, 0.9), mode=1, dtype=np.float32),
                RandFlip(prob=0.5, spatial_axis=[1, 2]),
                RandAxisFlip(prob=0.5),
                RandRotate90(prob=0, spatial_axes=(1, 2)),
                RandZoom(prob=0.5, min_zoom=0.5, max_zoom=1.1, keep_size=True),
                RandRotate(prob=0.5, range_x=np.pi, mode="bilinear", align_corners=True, dtype=np.float64),
                RandAffine(prob=0.5, rotate_range=np.pi, mode="nearest"),
                ResizeWithPadOrCrop(100),
                CastToType(dtype=torch.uint8),
            ]
        )

        # num workers = 0 for mac or gpu transforms
        num_workers = 0 if sys.platform != "linux" or torch.cuda.is_available() else 2
        dataset = Dataset(data, transform=transform)
        self.assertIsInstance(transform.inverse(dataset[0]), MetaTensor)
        loader = DataLoader(dataset, num_workers=num_workers, batch_size=1)
        inverter = Invert(transform=transform, nearest_interp=True, device="cpu", post_func=torch.as_tensor)

        for d in loader:
            d = decollate_batch(d)
            for item in d:
                orig = deepcopy(item)
                i = inverter(item)
                self.assertTupleEqual(orig.shape[1:], (100, 100, 100))
                # check the nearest interpolation mode
                assert_allclose(i.to(torch.uint8).to(torch.float), i.to(torch.float))
                self.assertTupleEqual(i.shape[1:], (101, 100, 107))
        # check labels match
        reverted = i.detach().cpu().numpy().astype(np.int32)
        original = LoadImage(image_only=True)(data[-1])
        n_good = np.sum(np.isclose(reverted, original.numpy(), atol=1e-3))
        reverted_name = i.meta["filename_or_obj"]
        original_name = original.meta["filename_or_obj"]
        self.assertEqual(reverted_name, original_name)
        print("invert diff", reverted.size - n_good)
        self.assertTrue((reverted.size - n_good) < 300000, f"diff. {reverted.size - n_good}")
        set_determinism(seed=None)

    def test_invert_warn_pending(self):
        # this test shouldn't raise a warning or error any more as that issue was fixed
        # by https://github.com/Project-MONAI/MONAI/pull/6257
        set_determinism(seed=0)
        im_fname = make_nifti_image(create_test_image_3d(101, 100, 107, noise_max=100)[1])  # label image, discrete
        transform = Compose(
            [LoadImage(image_only=True), EnsureChannelFirst(), Orientation("RPS"), Lambda(func=lambda x: x)], lazy=True
        )
        output = transform([im_fname for _ in range(2)])
        transform.inverse(output)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import unittest

import numpy as np
from parameterized import parameterized

from monai.transforms import get_extreme_points
from tests.test_utils import TEST_NDARRAYS

TESTS = []
for p in TEST_NDARRAYS:
    TESTS.append(
        [
            {
                "img": p(np.array([[0, 1, 0], [0, 0, 1], [0, 1, 0], [1, 0, 0]])),
                "rand_state": np.random,
                "background": 0,
                "pert": 0.0,
            },
            [(0, 1), (3, 0), (3, 0), (1, 2)],
        ]
    )

    TESTS.append(
        [
            {
                "img": p(np.array([[0, 1, 0], [1, 1, 1], [0, 1, 0], [0, 1, 0]])),
                "rand_state": np.random,
                "background": 0,
                "pert": 0.0,
            },
            [(0, 1), (3, 1), (1, 0), (1, 2)],
        ]
    )


class TestGetExtremePoints(unittest.TestCase):
    @parameterized.expand(TESTS)
    def test_type_shape(self, input_data, expected):
        result = get_extreme_points(**input_data)
        self.assertEqual(result, expected)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import os
import tempfile
import unittest
from urllib.error import HTTPError

from monai.apps.utils import download_url

YANDEX_MODEL_URL = (
    "https://cloud-api.yandex.net/v1/disk/public/resources/download?"
    "public_key=https%3A%2F%2Fdisk.yandex.ru%2Fd%2Fxs0gzlj2_irgWA"
)
YANDEX_MODEL_FLAWED_URL = (
    "https://cloud-api.yandex.net/v1/disk/public/resources/download?"
    "public_key=https%3A%2F%2Fdisk.yandex.ru%2Fd%2Fxs0gzlj2_irgWA-url-with-error"
)


class TestDownloadUrlYandex(unittest.TestCase):

    @unittest.skip("data source unstable")
    def test_verify(self):
        with tempfile.TemporaryDirectory() as tempdir:
            download_url(url=YANDEX_MODEL_URL, filepath=os.path.join(tempdir, "model.pt"))

    def test_verify_error(self):
        with tempfile.TemporaryDirectory() as tempdir:
            with self.assertRaises(HTTPError):
                download_url(url=YANDEX_MODEL_FLAWED_URL, filepath=os.path.join(tempdir, "model.pt"))


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import json
import os
import warnings
from collections.abc import Sequence
from pathlib import Path
from typing import overload

from monai.config import KeysCollection, PathLike
from monai.data.utils import partition_dataset, select_cross_validation_folds
from monai.utils import ensure_tuple


@overload
def _compute_path(base_dir: PathLike, element: PathLike, check_path: bool = False) -> str: ...


@overload
def _compute_path(base_dir: PathLike, element: list[PathLike], check_path: bool = False) -> list[str]: ...


def _compute_path(base_dir, element, check_path=False):
    """
    Args:
        base_dir: the base directory of the dataset.
        element: file path(s) to append to directory.
        check_path: if `True`, only compute when the result is an existing path.

    Raises:
        TypeError: When ``element`` contains a non ``str``.
        TypeError: When ``element`` type is not in ``Union[list, str]``.

    """

    def _join_path(base_dir: PathLike, item: PathLike):
        result = os.path.normpath(os.path.join(base_dir, item))
        if check_path and not os.path.exists(result):
            # if not an existing path, don't join with base dir
            return f"{item}"
        return f"{result}"

    if isinstance(element, (str, os.PathLike)):
        return _join_path(base_dir, element)
    if isinstance(element, list):
        for e in element:
            if not isinstance(e, (str, os.PathLike)):
                return element
        return [_join_path(base_dir, e) for e in element]
    return element


def _append_paths(base_dir: PathLike, is_segmentation: bool, items: list[dict]) -> list[dict]:
    """
    Args:
        base_dir: the base directory of the dataset.
        is_segmentation: whether the datalist is for segmentation task.
        items: list of data items, each of which is a dict keyed by element names.

    Raises:
        TypeError: When ``items`` contains a non ``dict``.

    """
    for item in items:
        if not isinstance(item, dict):
            raise TypeError(f"Every item in items must be a dict but got {type(item).__name__}.")
        for k, v in item.items():
            if k == "image" or is_segmentation and k == "label":
                item[k] = _compute_path(base_dir, v, check_path=False)
            else:
                # for other items, auto detect whether it's a valid path
                item[k] = _compute_path(base_dir, v, check_path=True)
    return items


def load_decathlon_datalist(
    data_list_file_path: PathLike,
    is_segmentation: bool = True,
    data_list_key: str = "training",
    base_dir: PathLike | None = None,
) -> list[dict]:
    """Load image/label paths of decathlon challenge from JSON file

    Json file is similar to what you get from http://medicaldecathlon.com/
    Those dataset.json files

    Args:
        data_list_file_path: the path to the json file of datalist.
        is_segmentation: whether the datalist is for segmentation task, default is True.
        data_list_key: the key to get a list of dictionary to be used, default is "training".
        base_dir: the base directory of the dataset, if None, use the datalist directory.

    Raises:
        ValueError: When ``data_list_file_path`` does not point to a file.
        ValueError: When ``data_list_key`` is not specified in the data list file.

    Returns a list of data items, each of which is a dict keyed by element names, for example:

    .. code-block::

        [
            {'image': '/workspace/data/chest_19.nii.gz',  'label': 0},
            {'image': '/workspace/data/chest_31.nii.gz',  'label': 1}
        ]

    """
    data_list_file_path = Path(data_list_file_path)
    if not data_list_file_path.is_file():
        raise ValueError(f"Data list file {data_list_file_path} does not exist.")
    with open(data_list_file_path) as json_file:
        json_data = json.load(json_file)
    if data_list_key not in json_data:
        raise ValueError(f'Data list {data_list_key} not specified in "{data_list_file_path}".')
    expected_data = json_data[data_list_key]
    if data_list_key == "test" and not isinstance(expected_data[0], dict):
        # decathlon datalist may save the test images in a list directly instead of dict
        expected_data = [{"image": i} for i in expected_data]

    if base_dir is None:
        base_dir = data_list_file_path.parent

    return _append_paths(base_dir, is_segmentation, expected_data)


def load_decathlon_properties(data_property_file_path: PathLike, property_keys: Sequence[str] | str) -> dict:
    """Load the properties from the JSON file contains data property with specified `property_keys`.

    Args:
        data_property_file_path: the path to the JSON file of data properties.
        property_keys: expected keys to load from the JSON file, for example, we have these keys
            in the decathlon challenge:
            `name`, `description`, `reference`, `licence`, `tensorImageSize`,
            `modality`, `labels`, `numTraining`, `numTest`, etc.

    """
    data_property_file_path = Path(data_property_file_path)
    if not data_property_file_path.is_file():
        raise ValueError(f"Data property file {data_property_file_path} does not exist.")
    with open(data_property_file_path) as json_file:
        json_data = json.load(json_file)

    properties = {}
    for key in ensure_tuple(property_keys):
        if key not in json_data:
            raise KeyError(f"key {key} is not in the data property file.")
        properties[key] = json_data[key]
    return properties


def check_missing_files(
    datalist: list[dict], keys: KeysCollection, root_dir: PathLike | None = None, allow_missing_keys: bool = False
):
    """Checks whether some files in the Decathlon datalist are missing.
    It would be helpful to check missing files before a heavy training run.

    Args:
        datalist: a list of data items, every item is a dictionary.
            usually generated by `load_decathlon_datalist` API.
        keys: expected keys to check in the datalist.
        root_dir: if not None, provides the root dir for the relative file paths in `datalist`.
        allow_missing_keys: whether allow missing keys in the datalist items.
            if False, raise exception if missing. default to False.

    Returns:
        A list of missing filenames.

    """
    missing_files = []
    for item in datalist:
        for k in ensure_tuple(keys):
            if k not in item:
                if not allow_missing_keys:
                    raise ValueError(f"key `{k}` is missing in the datalist item: {item}")
                continue

            for f in ensure_tuple(item[k]):
                if not isinstance(f, (str, os.PathLike)):
                    raise ValueError(f"filepath of key `{k}` must be a string or a list of strings, but got: {f}.")
                f = Path(f)
                if isinstance(root_dir, (str, os.PathLike)):
                    f = Path(root_dir).joinpath(f)
                if not f.exists():
                    missing_files.append(f)

    return missing_files


def create_cross_validation_datalist(
    datalist: list[dict],
    nfolds: int,
    train_folds: Sequence[int] | int,
    val_folds: Sequence[int] | int,
    train_key: str = "training",
    val_key: str = "validation",
    filename: Path | str | None = None,
    shuffle: bool = True,
    seed: int = 0,
    check_missing: bool = False,
    keys: KeysCollection | None = None,
    root_dir: str | None = None,
    allow_missing_keys: bool = False,
    raise_error: bool = True,
):
    """
    Utility to create new Decathlon style datalist based on cross validation partition.

    Args:
        datalist: loaded list of dictionaries for all the items to partition.
        nfolds: number of the kfold split.
        train_folds: indices of folds for training part.
        val_folds: indices of folds for validation part.
        train_key: the key of train part in the new datalist, defaults to "training".
        val_key: the key of validation part in the new datalist, defaults to "validation".
        filename: if not None and ends with ".json", save the new datalist into JSON file.
        shuffle: whether to shuffle the datalist before partition, defaults to `True`.
        seed: if `shuffle` is True, set the random seed, defaults to `0`.
        check_missing: whether to check all the files specified by `keys` are existing.
        keys: if not None and check_missing_files is True, the expected keys to check in the datalist.
        root_dir: if not None, provides the root dir for the relative file paths in `datalist`.
        allow_missing_keys: if check_missing_files is `True`, whether allow missing keys in the datalist items.
            if False, raise exception if missing. default to False.
        raise_error: when found missing files, if `True`, raise exception and stop, if `False`, print warning.

    """
    if check_missing and keys is not None:
        files = check_missing_files(datalist, keys, root_dir, allow_missing_keys)
        if files:
            msg = f"some files of the datalist are missing: {files}"
            if raise_error:
                raise ValueError(msg)
            warnings.warn(msg)

    data = partition_dataset(data=datalist, num_partitions=nfolds, shuffle=shuffle, seed=seed)
    train_list = select_cross_validation_folds(partitions=data, folds=train_folds)
    val_list = select_cross_validation_folds(partitions=data, folds=val_folds)
    ret = {train_key: train_list, val_key: val_list}
    if isinstance(filename, (str, Path)):
        with open(filename, "w") as f:
            json.dump(ret, f, indent=4)

    return ret



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import unittest

from monai.utils.module import get_package_version


class TestGetVersion(unittest.TestCase):

    def test_default(self):
        output = get_package_version("42foobarnoexist")
        self.assertIn("UNKNOWN", output)

        output = get_package_version("numpy")
        self.assertNotIn("UNKNOWN", output)

    def test_msg(self):
        output = get_package_version("42foobarnoexist", "test")
        self.assertIn("test", output)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import unittest
from typing import TYPE_CHECKING
from unittest import skipUnless

import torch
from parameterized import parameterized

from monai.networks import eval_mode
from monai.networks.nets import DenseNet121, Densenet169, DenseNet264, densenet201
from monai.utils import optional_import
from tests.test_utils import skip_if_downloading_fails, skip_if_quick, test_script_save

if TYPE_CHECKING:
    import torchvision

    has_torchvision = True
else:
    torchvision, has_torchvision = optional_import("torchvision")

device = "cuda" if torch.cuda.is_available() else "cpu"

TEST_CASE_1 = [  # 4-channel 3D, batch 2
    {"pretrained": False, "spatial_dims": 3, "in_channels": 2, "out_channels": 3, "norm": ("instance", {"eps": 1e-5})},
    (2, 2, 32, 64, 48),
    (2, 3),
]

TEST_CASE_2 = [  # 4-channel 2D, batch 2
    {"pretrained": False, "spatial_dims": 2, "in_channels": 2, "out_channels": 3, "act": "PRELU"},
    (2, 2, 32, 64),
    (2, 3),
]

TEST_CASE_3 = [  # 4-channel 1D, batch 1
    {"pretrained": False, "spatial_dims": 1, "in_channels": 2, "out_channels": 3},
    (1, 2, 32),
    (1, 3),
]

TEST_CASES = []
for case in [TEST_CASE_1, TEST_CASE_2, TEST_CASE_3]:
    for model in [DenseNet121, Densenet169, densenet201, DenseNet264]:
        TEST_CASES.append([model, *case])

TEST_SCRIPT_CASES = [[model, *TEST_CASE_1] for model in [DenseNet121, Densenet169, densenet201, DenseNet264]]

TEST_PRETRAINED_2D_CASE_1 = [  # 4-channel 2D, batch 2
    DenseNet121,
    {"pretrained": True, "progress": True, "spatial_dims": 2, "in_channels": 2, "out_channels": 3},
    (1, 2, 32, 64),
    (1, 3),
]

TEST_PRETRAINED_2D_CASE_2 = [  # 4-channel 2D, batch 2
    DenseNet121,
    {"pretrained": True, "progress": False, "spatial_dims": 2, "in_channels": 2, "out_channels": 1},
    (1, 2, 32, 64),
    (1, 1),
]

TEST_PRETRAINED_2D_CASE_3 = [
    DenseNet121,
    {"pretrained": True, "progress": False, "spatial_dims": 2, "in_channels": 3, "out_channels": 1},
    (1, 3, 32, 32),
]


class TestPretrainedDENSENET(unittest.TestCase):
    @parameterized.expand([TEST_PRETRAINED_2D_CASE_1, TEST_PRETRAINED_2D_CASE_2])
    @skip_if_quick
    def test_121_2d_shape_pretrain(self, model, input_param, input_shape, expected_shape):
        with skip_if_downloading_fails():
            net = model(**input_param).to(device)
        with eval_mode(net):
            result = net.forward(torch.randn(input_shape).to(device))
            self.assertEqual(result.shape, expected_shape)

    @parameterized.expand([TEST_PRETRAINED_2D_CASE_3])
    @skipUnless(has_torchvision, "Requires `torchvision` package.")
    def test_pretrain_consistency(self, model, input_param, input_shape):
        example = torch.randn(input_shape).to(device)
        with skip_if_downloading_fails():
            net = model(**input_param).to(device)
        with eval_mode(net):
            result = net.features.forward(example)
        torchvision_net = torchvision.models.densenet121(pretrained=True).to(device)
        with eval_mode(torchvision_net):
            expected_result = torchvision_net.features.forward(example)
        self.assertTrue(torch.all(result == expected_result))


class TestDENSENET(unittest.TestCase):
    @parameterized.expand(TEST_CASES)
    def test_densenet_shape(self, model, input_param, input_shape, expected_shape):
        net = model(**input_param).to(device)
        with eval_mode(net):
            result = net.forward(torch.randn(input_shape).to(device))
            self.assertEqual(result.shape, expected_shape)

    @parameterized.expand(TEST_SCRIPT_CASES)
    def test_script(self, model, input_param, input_shape, expected_shape):
        net = model(**input_param)
        test_data = torch.randn(input_shape)
        test_script_save(net, test_data)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import unittest

import torch

from monai.networks.layers import ConjugateGradient


class TestConjugateGradient(unittest.TestCase):

    def test_real_valued_inverse(self):
        """Test ConjugateGradient with real-valued input: when the input is real
        value, the output should be the inverse of the matrix."""
        a_dim = 3
        a_mat = torch.tensor([[1, 2, 3], [2, 1, 2], [3, 2, 1]], dtype=torch.float)

        def a_op(x):
            return a_mat @ x

        cg_solver = ConjugateGradient(a_op, num_iter=100)
        # define the measurement
        y = torch.tensor([1, 2, 3], dtype=torch.float)
        # solve for x
        x = cg_solver(torch.zeros(a_dim), y)
        x_ref = torch.linalg.solve(a_mat, y)
        # assert torch.allclose(x, x_ref, atol=1e-6), 'CG solver failed to converge to reference solution'
        self.assertTrue(torch.allclose(x, x_ref, atol=1e-6))

    def test_complex_valued_inverse(self):
        a_dim = 3
        a_mat = torch.tensor([[1, 2, 3], [2, 1, 2], [3, 2, 1]], dtype=torch.complex64)

        def a_op(x):
            return a_mat @ x

        cg_solver = ConjugateGradient(a_op, num_iter=100)
        y = torch.tensor([1, 2, 3], dtype=torch.complex64)
        x = cg_solver(torch.zeros(a_dim, dtype=torch.complex64), y)
        x_ref = torch.linalg.solve(a_mat, y)
        self.assertTrue(torch.allclose(x, x_ref, atol=1e-6))


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from monai.config import KeysCollection
from monai.transforms.traits import InvertibleTrait
from monai.transforms.transform import MapTransform

__all__ = ["ApplyPendingd", "ApplyPendingD", "ApplyPendingDict"]


class ApplyPendingd(InvertibleTrait, MapTransform):
    """
    ApplyPendingd can be inserted into a pipeline that is being executed lazily in order
    to ensure resampling happens before the next transform. It doesn't do anything itself,
    but its presence causes the pipeline to be executed as it doesn't implement ``LazyTrait``

    See ``Compose`` for a detailed explanation of the lazy resampling feature.

    Args:
        keys: the keys for tensors that should have their pending transforms executed
    """

    def __init__(self, keys: KeysCollection):
        super().__init__(keys)

    def __call__(self, data):
        return data

    def inverse(self, data):
        return data


ApplyPendingD = ApplyPendingDict = ApplyPendingd



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import unittest

import torch
from parameterized import parameterized

from monai.networks import eval_mode
from monai.networks.nets.vqvae import VQVAE
from tests.test_utils import SkipIfBeforePyTorchVersion

TEST_CASES = [
    [
        {
            "spatial_dims": 2,
            "in_channels": 1,
            "out_channels": 1,
            "channels": (4, 4),
            "num_res_layers": 1,
            "num_res_channels": (4, 4),
            "downsample_parameters": ((2, 4, 1, 1),) * 2,
            "upsample_parameters": ((2, 4, 1, 1, 0),) * 2,
            "num_embeddings": 8,
            "embedding_dim": 8,
        },
        (1, 1, 8, 8),
        (1, 1, 8, 8),
    ],
    [
        {
            "spatial_dims": 3,
            "in_channels": 1,
            "out_channels": 1,
            "channels": (4, 4),
            "num_res_layers": 1,
            "num_res_channels": 4,
            "downsample_parameters": ((2, 4, 1, 1),) * 2,
            "upsample_parameters": ((2, 4, 1, 1, 0),) * 2,
            "num_embeddings": 8,
            "embedding_dim": 8,
        },
        (1, 1, 8, 8, 8),
        (1, 1, 8, 8, 8),
    ],
    [
        {
            "spatial_dims": 2,
            "in_channels": 1,
            "out_channels": 1,
            "channels": (4, 4),
            "num_res_layers": 1,
            "num_res_channels": (4, 4),
            "downsample_parameters": (2, 4, 1, 1),
            "upsample_parameters": ((2, 4, 1, 1, 0),) * 2,
            "num_embeddings": 8,
            "embedding_dim": 8,
        },
        (1, 1, 8, 8),
        (1, 1, 8, 8),
    ],
    [
        {
            "spatial_dims": 3,
            "in_channels": 1,
            "out_channels": 1,
            "channels": (4, 4),
            "num_res_layers": 1,
            "num_res_channels": (4, 4),
            "downsample_parameters": ((2, 4, 1, 1),) * 2,
            "upsample_parameters": (2, 4, 1, 1, 0),
            "num_embeddings": 8,
            "embedding_dim": 8,
        },
        (1, 1, 8, 8, 8),
        (1, 1, 8, 8, 8),
    ],
]

TEST_LATENT_SHAPE = {
    "spatial_dims": 2,
    "in_channels": 1,
    "out_channels": 1,
    "downsample_parameters": ((2, 4, 1, 1),) * 2,
    "upsample_parameters": ((2, 4, 1, 1, 0),) * 2,
    "num_res_layers": 1,
    "channels": (8, 8),
    "num_res_channels": (8, 8),
    "num_embeddings": 16,
    "embedding_dim": 8,
}


class TestVQVAE(unittest.TestCase):
    @parameterized.expand(TEST_CASES)
    def test_shape(self, input_param, input_shape, expected_shape):
        device = "cuda" if torch.cuda.is_available() else "cpu"

        net = VQVAE(**input_param).to(device)

        with eval_mode(net):
            result, _ = net(torch.randn(input_shape).to(device))

        self.assertEqual(result.shape, expected_shape)

    @parameterized.expand(TEST_CASES)
    @SkipIfBeforePyTorchVersion((1, 11))
    def test_shape_with_checkpoint(self, input_param, input_shape, expected_shape):
        device = "cuda" if torch.cuda.is_available() else "cpu"
        input_param = input_param.copy()
        input_param.update({"use_checkpointing": True})

        net = VQVAE(**input_param).to(device)

        with eval_mode(net):
            result, _ = net(torch.randn(input_shape).to(device))

        self.assertEqual(result.shape, expected_shape)

    # Removed this test case since TorchScript currently does not support activation checkpoint.
    # def test_script(self):
    #     net = VQVAE(
    #         spatial_dims=2,
    #         in_channels=1,
    #         out_channels=1,
    #         downsample_parameters=((2, 4, 1, 1),) * 2,
    #         upsample_parameters=((2, 4, 1, 1, 0),) * 2,
    #         num_res_layers=1,
    #         channels=(8, 8),
    #         num_res_channels=(8, 8),
    #         num_embeddings=16,
    #         embedding_dim=8,
    #         ddp_sync=False,
    #     )
    #     test_data = torch.randn(1, 1, 16, 16)
    #     test_script_save(net, test_data)

    def test_channels_not_same_size_of_num_res_channels(self):
        with self.assertRaises(ValueError):
            VQVAE(
                spatial_dims=2,
                in_channels=1,
                out_channels=1,
                channels=(16, 16),
                num_res_channels=(16, 16, 16),
                downsample_parameters=((2, 4, 1, 1),) * 2,
                upsample_parameters=((2, 4, 1, 1, 0),) * 2,
            )

    def test_channels_not_same_size_of_downsample_parameters(self):
        with self.assertRaises(ValueError):
            VQVAE(
                spatial_dims=2,
                in_channels=1,
                out_channels=1,
                channels=(16, 16),
                num_res_channels=(16, 16),
                downsample_parameters=((2, 4, 1, 1),) * 3,
                upsample_parameters=((2, 4, 1, 1, 0),) * 2,
            )

    def test_channels_not_same_size_of_upsample_parameters(self):
        with self.assertRaises(ValueError):
            VQVAE(
                spatial_dims=2,
                in_channels=1,
                out_channels=1,
                channels=(16, 16),
                num_res_channels=(16, 16),
                downsample_parameters=((2, 4, 1, 1),) * 2,
                upsample_parameters=((2, 4, 1, 1, 0),) * 3,
            )

    def test_downsample_parameters_not_sequence_or_int(self):
        with self.assertRaises(ValueError):
            VQVAE(
                spatial_dims=2,
                in_channels=1,
                out_channels=1,
                channels=(16, 16),
                num_res_channels=(16, 16),
                downsample_parameters=(("test", 4, 1, 1),) * 2,
                upsample_parameters=((2, 4, 1, 1, 0),) * 2,
            )

    def test_upsample_parameters_not_sequence_or_int(self):
        with self.assertRaises(ValueError):
            VQVAE(
                spatial_dims=2,
                in_channels=1,
                out_channels=1,
                channels=(16, 16),
                num_res_channels=(16, 16),
                downsample_parameters=((2, 4, 1, 1),) * 2,
                upsample_parameters=(("test", 4, 1, 1, 0),) * 2,
            )

    def test_downsample_parameter_length_different_4(self):
        with self.assertRaises(ValueError):
            VQVAE(
                spatial_dims=2,
                in_channels=1,
                out_channels=1,
                channels=(16, 16),
                num_res_channels=(16, 16),
                downsample_parameters=((2, 4, 1),) * 3,
                upsample_parameters=((2, 4, 1, 1, 0),) * 2,
            )

    def test_upsample_parameter_length_different_5(self):
        with self.assertRaises(ValueError):
            VQVAE(
                spatial_dims=2,
                in_channels=1,
                out_channels=1,
                channels=(16, 16),
                num_res_channels=(16, 16, 16),
                downsample_parameters=((2, 4, 1, 1),) * 2,
                upsample_parameters=((2, 4, 1, 1, 0, 1),) * 3,
            )

    def test_encode_shape(self):
        device = "cuda" if torch.cuda.is_available() else "cpu"

        net = VQVAE(**TEST_LATENT_SHAPE).to(device)

        with eval_mode(net):
            latent = net.encode(torch.randn(1, 1, 32, 32).to(device))

        self.assertEqual(latent.shape, (1, 8, 8, 8))

    def test_index_quantize_shape(self):
        device = "cuda" if torch.cuda.is_available() else "cpu"

        net = VQVAE(**TEST_LATENT_SHAPE).to(device)

        with eval_mode(net):
            latent = net.index_quantize(torch.randn(1, 1, 32, 32).to(device))

        self.assertEqual(latent.shape, (1, 8, 8))

    def test_decode_shape(self):
        device = "cuda" if torch.cuda.is_available() else "cpu"

        net = VQVAE(**TEST_LATENT_SHAPE).to(device)

        with eval_mode(net):
            latent = net.decode(torch.randn(1, 8, 8, 8).to(device))

        self.assertEqual(latent.shape, (1, 1, 32, 32))

    def test_decode_samples_shape(self):
        device = "cuda" if torch.cuda.is_available() else "cpu"

        net = VQVAE(**TEST_LATENT_SHAPE).to(device)

        with eval_mode(net):
            latent = net.decode_samples(torch.randint(low=0, high=16, size=(1, 8, 8)).to(device))

        self.assertEqual(latent.shape, (1, 1, 32, 32))


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import torch
import torch.nn as nn

from monai.networks.blocks.cablock import CABlock, FeedForward
from monai.networks.blocks.convolutions import Convolution
from monai.networks.blocks.downsample import DownSample
from monai.networks.blocks.upsample import UpSample
from monai.networks.layers.factories import Norm
from monai.utils.enums import DownsampleMode, UpsampleMode


class MDTATransformerBlock(nn.Module):
    """Basic transformer unit combining MDTA and GDFN with skip connections.
    Unlike standard transformers that use LayerNorm, this block uses Instance Norm
    for better adaptation to image restoration tasks.

    Args:
        spatial_dims: Number of spatial dimensions (2D or 3D)
        dim: Number of input channels
        num_heads: Number of attention heads
        ffn_expansion_factor: Expansion factor for feed-forward network
        bias: Whether to use bias in attention layers
        layer_norm_use_bias: Whether to use bias in layer normalization. Defaults to False.
        flash_attention: Whether to use flash attention optimization. Defaults to False.
    """

    def __init__(
        self,
        spatial_dims: int,
        dim: int,
        num_heads: int,
        ffn_expansion_factor: float,
        bias: bool,
        layer_norm_use_bias: bool = False,
        flash_attention: bool = False,
    ):
        super().__init__()
        self.norm1 = Norm[Norm.INSTANCE, spatial_dims](dim, affine=layer_norm_use_bias)
        self.attn = CABlock(spatial_dims, dim, num_heads, bias, flash_attention)
        self.norm2 = Norm[Norm.INSTANCE, spatial_dims](dim, affine=layer_norm_use_bias)
        self.ffn = FeedForward(spatial_dims, dim, ffn_expansion_factor, bias)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = x + self.attn(self.norm1(x))
        x = x + self.ffn(self.norm2(x))
        return x


class OverlapPatchEmbed(Convolution):
    """Initial feature extraction using overlapped convolutions.
    Unlike standard patch embeddings that use non-overlapping patches,
    this approach maintains spatial continuity through 3x3 convolutions.

    Args:
        spatial_dims: Number of spatial dimensions (2D or 3D)
        in_channels: Number of input channels
        embed_dim: Dimension of embedded features. Defaults to 48.
        bias: Whether to use bias in convolution layer. Defaults to False.
    """

    def __init__(self, spatial_dims: int, in_channels: int = 3, embed_dim: int = 48, bias: bool = False):
        super().__init__(
            spatial_dims=spatial_dims,
            in_channels=in_channels,
            out_channels=embed_dim,
            kernel_size=3,
            strides=1,
            padding=1,
            bias=bias,
            conv_only=True,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = super().forward(x)
        return x


class Restormer(nn.Module):
    """Restormer: Efficient Transformer for High-Resolution Image Restoration.

    Implements a U-Net style architecture with transformer blocks, combining:
    - Multi-scale feature processing through progressive down/upsampling
    - Efficient attention via MDTA blocks
    - Local feature mixing through GDFN
    - Skip connections for preserving spatial details

    Architecture:
        - Encoder: Progressive feature downsampling with increasing channels
        - Latent: Deep feature processing at lowest resolution
        - Decoder: Progressive upsampling with skip connections
        - Refinement: Final feature enhancement
    """

    def __init__(
        self,
        spatial_dims: int = 2,
        in_channels: int = 3,
        out_channels: int = 3,
        dim: int = 48,
        num_blocks: tuple[int, ...] = (1, 1, 1, 1),
        heads: tuple[int, ...] = (1, 1, 1, 1),
        num_refinement_blocks: int = 4,
        ffn_expansion_factor: float = 2.66,
        bias: bool = False,
        layer_norm_use_bias: bool = True,
        dual_pixel_task: bool = False,
        flash_attention: bool = False,
    ) -> None:
        super().__init__()
        """Initialize Restormer model.

        Args:
            spatial_dims: Number of spatial dimensions (2D or 3D)
            in_channels: Number of input image channels
            out_channels: Number of output image channels
            dim: Base feature dimension. Defaults to 48.
            num_blocks: Number of transformer blocks at each scale. Defaults to (1,1,1,1).
            heads: Number of attention heads at each scale. Defaults to (1,1,1,1).
            num_refinement_blocks: Number of final refinement blocks. Defaults to 4.
            ffn_expansion_factor: Expansion factor for feed-forward network. Defaults to 2.66.
            bias: Whether to use bias in convolutions. Defaults to False.
            layer_norm_use_bias: Whether to use bias in layer normalization. Defaults to True.
            dual_pixel_task: Enable dual-pixel specific processing. Defaults to False.
            flash_attention: Use flash attention if available. Defaults to False.

        Note:
            The number of blocks must be greater than 1
            The length of num_blocks and heads must be equal
            All values in num_blocks must be greater than 0
        """
        # Check input parameters
        assert len(num_blocks) > 1, "Number of blocks must be greater than 1"
        assert len(num_blocks) == len(heads), "Number of blocks and heads must be equal"
        assert all(n > 0 for n in num_blocks), "Number of blocks must be greater than 0"

        # Initial feature extraction
        self.patch_embed = OverlapPatchEmbed(spatial_dims, in_channels, dim)
        self.encoder_levels = nn.ModuleList()
        self.downsamples = nn.ModuleList()
        self.decoder_levels = nn.ModuleList()
        self.upsamples = nn.ModuleList()
        self.reduce_channels = nn.ModuleList()
        num_steps = len(num_blocks) - 1
        self.num_steps = num_steps
        self.spatial_dims = spatial_dims
        spatial_multiplier = 2 ** (spatial_dims - 1)

        # Define encoder levels
        for n in range(num_steps):
            current_dim = dim * (2) ** (n)
            next_dim = current_dim // spatial_multiplier
            self.encoder_levels.append(
                nn.Sequential(
                    *[
                        MDTATransformerBlock(
                            spatial_dims=spatial_dims,
                            dim=current_dim,
                            num_heads=heads[n],
                            ffn_expansion_factor=ffn_expansion_factor,
                            bias=bias,
                            layer_norm_use_bias=layer_norm_use_bias,
                            flash_attention=flash_attention,
                        )
                        for _ in range(num_blocks[n])
                    ]
                )
            )

            self.downsamples.append(
                DownSample(
                    spatial_dims=self.spatial_dims,
                    in_channels=current_dim,
                    out_channels=next_dim,
                    mode=DownsampleMode.PIXELUNSHUFFLE,
                    scale_factor=2,
                    bias=bias,
                )
            )

        # Define latent space
        latent_dim = dim * (2) ** (num_steps)
        self.latent = nn.Sequential(
            *[
                MDTATransformerBlock(
                    spatial_dims=spatial_dims,
                    dim=latent_dim,
                    num_heads=heads[num_steps],
                    ffn_expansion_factor=ffn_expansion_factor,
                    bias=bias,
                    layer_norm_use_bias=layer_norm_use_bias,
                    flash_attention=flash_attention,
                )
                for _ in range(num_blocks[num_steps])
            ]
        )

        # Define decoder levels
        for n in reversed(range(num_steps)):
            current_dim = dim * (2) ** (n)
            next_dim = dim * (2) ** (n + 1)
            self.upsamples.append(
                UpSample(
                    spatial_dims=self.spatial_dims,
                    in_channels=next_dim,
                    out_channels=(current_dim),
                    mode=UpsampleMode.PIXELSHUFFLE,
                    scale_factor=2,
                    bias=bias,
                    apply_pad_pool=False,
                )
            )

            # Reduce channel layers to deal with skip connections
            if n != 0:
                self.reduce_channels.append(
                    Convolution(
                        spatial_dims=self.spatial_dims,
                        in_channels=next_dim,
                        out_channels=current_dim,
                        kernel_size=1,
                        bias=bias,
                        conv_only=True,
                    )
                )
                decoder_dim = current_dim
            else:
                decoder_dim = next_dim

            self.decoder_levels.append(
                nn.Sequential(
                    *[
                        MDTATransformerBlock(
                            spatial_dims=spatial_dims,
                            dim=decoder_dim,
                            num_heads=heads[n],
                            ffn_expansion_factor=ffn_expansion_factor,
                            bias=bias,
                            layer_norm_use_bias=layer_norm_use_bias,
                            flash_attention=flash_attention,
                        )
                        for _ in range(num_blocks[n])
                    ]
                )
            )

        # Final refinement and output
        self.refinement = nn.Sequential(
            *[
                MDTATransformerBlock(
                    spatial_dims=spatial_dims,
                    dim=decoder_dim,
                    num_heads=heads[0],
                    ffn_expansion_factor=ffn_expansion_factor,
                    bias=bias,
                    layer_norm_use_bias=layer_norm_use_bias,
                    flash_attention=flash_attention,
                )
                for _ in range(num_refinement_blocks)
            ]
        )
        self.dual_pixel_task = dual_pixel_task
        if self.dual_pixel_task:
            self.skip_conv = Convolution(
                spatial_dims=self.spatial_dims,
                in_channels=dim,
                out_channels=dim * 2,
                kernel_size=1,
                bias=bias,
                conv_only=True,
            )
        self.output = Convolution(
            spatial_dims=self.spatial_dims,
            in_channels=dim * 2,
            out_channels=out_channels,
            kernel_size=3,
            strides=1,
            padding=1,
            bias=bias,
            conv_only=True,
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass of Restormer.
        Processes input through encoder-decoder architecture with skip connections.
        Args:
            inp_img: Input image tensor of shape (B, C, H, W, [D])

        Returns:
            Restored image tensor of shape (B, C, H, W, [D])
        """
        assert all(
            x.shape[-i] > 2**self.num_steps for i in range(1, self.spatial_dims + 1)
        ), "All spatial dimensions should be larger than 2^number_of_step"

        # Patch embedding
        x = self.patch_embed(x)
        skip_connections = []

        # Encoding path
        for _idx, (encoder, downsample) in enumerate(zip(self.encoder_levels, self.downsamples)):
            x = encoder(x)
            skip_connections.append(x)
            x = downsample(x)

        # Latent space
        x = self.latent(x)

        # Decoding path
        for idx in range(len(self.decoder_levels)):
            x = self.upsamples[idx](x)
            x = torch.concat([x, skip_connections[-(idx + 1)]], 1)
            if idx < len(self.decoder_levels) - 1:
                x = self.reduce_channels[idx](x)
            x = self.decoder_levels[idx](x)

        # Final refinement
        x = self.refinement(x)

        if self.dual_pixel_task:
            x = x + self.skip_conv(skip_connections[0])
            x = self.output(x)
        else:
            x = self.output(x)

        return x



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Collection of the remote MMAR descriptors

See Also:
    - https://docs.nvidia.com/clara/clara-train-sdk/pt/mmar.html
"""

from __future__ import annotations

import os
from typing import Any

__all__ = ["MODEL_DESC", "RemoteMMARKeys"]


class RemoteMMARKeys:
    """
    Data keys used for loading MMAR.
    ID must uniquely define an MMAR.
    """

    ID = "id"  # unique MMAR
    NAME = "name"  # MMAR name for readability
    URL = "url"  # remote location of the MMAR, see also: `monai.apps.mmars.mmars._get_ngc_url`
    DOC = "doc"  # documentation page of the remote model, see also: `monai.apps.mmars.mmars._get_ngc_doc_url`
    FILE_TYPE = "file_type"  # type of the compressed MMAR
    HASH_TYPE = "hash_type"  # hashing method for the compressed MMAR
    HASH_VAL = "hash_val"  # hashing value for the compressed MMAR
    MODEL_FILE = "model_file"  # within an MMAR folder, the relative path to the model file
    CONFIG_FILE = "config_file"  # within an MMAR folder, the relative path to the config file (for model config)
    VERSION = "version"  # version of the MMAR


MODEL_DESC: tuple[dict[Any, Any], ...] = (
    {
        RemoteMMARKeys.ID: "clara_pt_spleen_ct_segmentation_1",
        RemoteMMARKeys.NAME: "clara_pt_spleen_ct_segmentation",
        RemoteMMARKeys.FILE_TYPE: "zip",
        RemoteMMARKeys.HASH_TYPE: "md5",
        RemoteMMARKeys.HASH_VAL: None,
        RemoteMMARKeys.MODEL_FILE: os.path.join("models", "model.pt"),
        RemoteMMARKeys.CONFIG_FILE: os.path.join("config", "config_train.json"),
        RemoteMMARKeys.VERSION: 1,
    },
    {
        RemoteMMARKeys.ID: "clara_pt_prostate_mri_segmentation_1",
        RemoteMMARKeys.NAME: "clara_pt_prostate_mri_segmentation",
        RemoteMMARKeys.FILE_TYPE: "zip",
        RemoteMMARKeys.HASH_TYPE: "md5",
        RemoteMMARKeys.HASH_VAL: None,
        RemoteMMARKeys.MODEL_FILE: os.path.join("models", "model.pt"),
        RemoteMMARKeys.CONFIG_FILE: os.path.join("config", "config_train.json"),
        RemoteMMARKeys.VERSION: 1,
    },
    {
        RemoteMMARKeys.ID: "clara_pt_covid19_ct_lesion_segmentation_1",
        RemoteMMARKeys.NAME: "clara_pt_covid19_ct_lesion_segmentation",
        RemoteMMARKeys.FILE_TYPE: "zip",
        RemoteMMARKeys.HASH_TYPE: "md5",
        RemoteMMARKeys.HASH_VAL: None,
        RemoteMMARKeys.MODEL_FILE: os.path.join("models", "model.pt"),
        RemoteMMARKeys.VERSION: 1,
    },
    {
        RemoteMMARKeys.ID: "clara_pt_covid19_3d_ct_classification_1",
        RemoteMMARKeys.NAME: "clara_pt_covid19_3d_ct_classification",
        RemoteMMARKeys.FILE_TYPE: "zip",
        RemoteMMARKeys.HASH_TYPE: "md5",
        RemoteMMARKeys.HASH_VAL: None,
        RemoteMMARKeys.MODEL_FILE: os.path.join("models", "model.pt"),
        RemoteMMARKeys.CONFIG_FILE: os.path.join("config", "config_train.json"),
        RemoteMMARKeys.VERSION: 1,
    },
    {
        RemoteMMARKeys.ID: "clara_pt_covid19_ct_lung_annotation_1",
        RemoteMMARKeys.NAME: "clara_pt_covid19_ct_lung_annotation",
        RemoteMMARKeys.FILE_TYPE: "zip",
        RemoteMMARKeys.HASH_TYPE: "md5",
        RemoteMMARKeys.HASH_VAL: None,
        RemoteMMARKeys.MODEL_FILE: os.path.join("models", "model.pt"),
        RemoteMMARKeys.VERSION: 1,
    },
    {
        RemoteMMARKeys.ID: "clara_pt_fed_learning_brain_tumor_mri_segmentation_1",
        RemoteMMARKeys.NAME: "clara_pt_fed_learning_brain_tumor_mri_segmentation",
        RemoteMMARKeys.FILE_TYPE: "zip",
        RemoteMMARKeys.HASH_TYPE: "md5",
        RemoteMMARKeys.HASH_VAL: None,
        RemoteMMARKeys.MODEL_FILE: os.path.join("models", "server", "best_FL_global_model.pt"),
        RemoteMMARKeys.VERSION: 1,
    },
    {
        RemoteMMARKeys.ID: "clara_pt_pathology_metastasis_detection_1",
        RemoteMMARKeys.NAME: "clara_pt_pathology_metastasis_detection",
        RemoteMMARKeys.FILE_TYPE: "zip",
        RemoteMMARKeys.HASH_TYPE: "md5",
        RemoteMMARKeys.HASH_VAL: None,
        RemoteMMARKeys.MODEL_FILE: os.path.join("models", "model.pt"),
        RemoteMMARKeys.CONFIG_FILE: os.path.join("config", "config_train.json"),
        RemoteMMARKeys.VERSION: 1,
    },
    {
        RemoteMMARKeys.ID: "clara_pt_brain_mri_segmentation_1",
        RemoteMMARKeys.NAME: "clara_pt_brain_mri_segmentation",
        RemoteMMARKeys.FILE_TYPE: "zip",
        RemoteMMARKeys.HASH_TYPE: "md5",
        RemoteMMARKeys.HASH_VAL: None,
        RemoteMMARKeys.MODEL_FILE: os.path.join("models", "model.pt"),
        RemoteMMARKeys.VERSION: 1,
    },
    {
        RemoteMMARKeys.ID: "clara_pt_brain_mri_segmentation_t1c_1",
        RemoteMMARKeys.NAME: "clara_pt_brain_mri_segmentation_t1c",
        RemoteMMARKeys.FILE_TYPE: "zip",
        RemoteMMARKeys.HASH_TYPE: "md5",
        RemoteMMARKeys.HASH_VAL: None,
        RemoteMMARKeys.MODEL_FILE: os.path.join("models", "model.pt"),
        RemoteMMARKeys.VERSION: 1,
    },
    {
        RemoteMMARKeys.ID: "clara_pt_liver_and_tumor_ct_segmentation_1",
        RemoteMMARKeys.NAME: "clara_pt_liver_and_tumor_ct_segmentation",
        RemoteMMARKeys.FILE_TYPE: "zip",
        RemoteMMARKeys.HASH_TYPE: "md5",
        RemoteMMARKeys.HASH_VAL: None,
        RemoteMMARKeys.MODEL_FILE: os.path.join("models", "model.pt"),
        RemoteMMARKeys.CONFIG_FILE: os.path.join("config", "config_train.json"),
        RemoteMMARKeys.VERSION: 1,
    },
    {
        RemoteMMARKeys.ID: "clara_pt_pancreas_and_tumor_ct_segmentation_1",
        RemoteMMARKeys.NAME: "clara_pt_pancreas_and_tumor_ct_segmentation",
        RemoteMMARKeys.FILE_TYPE: "zip",
        RemoteMMARKeys.HASH_TYPE: "md5",
        RemoteMMARKeys.HASH_VAL: None,
        RemoteMMARKeys.MODEL_FILE: os.path.join("models", "model.pt"),
        RemoteMMARKeys.CONFIG_FILE: os.path.join("config", "config_train.json"),
        RemoteMMARKeys.VERSION: 1,
    },
    {
        RemoteMMARKeys.ID: "clara_pt_brain_mri_annotation_t1c_1",
        RemoteMMARKeys.NAME: "clara_pt_brain_mri_annotation_t1c",
        RemoteMMARKeys.FILE_TYPE: "zip",
        RemoteMMARKeys.HASH_TYPE: "md5",
        RemoteMMARKeys.HASH_VAL: None,
        RemoteMMARKeys.MODEL_FILE: os.path.join("models", "model.pt"),
        RemoteMMARKeys.VERSION: 1,
    },
    {
        RemoteMMARKeys.ID: "clara_pt_spleen_ct_annotation_1",
        RemoteMMARKeys.NAME: "clara_pt_spleen_ct_annotation",
        RemoteMMARKeys.FILE_TYPE: "zip",
        RemoteMMARKeys.HASH_TYPE: "md5",
        RemoteMMARKeys.HASH_VAL: None,
        RemoteMMARKeys.MODEL_FILE: os.path.join("models", "model.pt"),
        RemoteMMARKeys.CONFIG_FILE: os.path.join("config", "config_train.json"),
        RemoteMMARKeys.VERSION: 1,
    },
    {
        RemoteMMARKeys.ID: "clara_pt_deepgrow_3d_annotation_1",
        RemoteMMARKeys.NAME: "clara_pt_deepgrow_3d_annotation",
        RemoteMMARKeys.FILE_TYPE: "zip",
        RemoteMMARKeys.HASH_TYPE: "md5",
        RemoteMMARKeys.HASH_VAL: None,
        RemoteMMARKeys.MODEL_FILE: os.path.join("models", "model.pt"),
        RemoteMMARKeys.CONFIG_FILE: os.path.join("config", "config_train.json"),
        RemoteMMARKeys.VERSION: 1,
    },
    {
        RemoteMMARKeys.ID: "clara_pt_deepgrow_2d_annotation_1",
        RemoteMMARKeys.NAME: "clara_pt_deepgrow_2d_annotation",
        RemoteMMARKeys.FILE_TYPE: "zip",
        RemoteMMARKeys.HASH_TYPE: "md5",
        RemoteMMARKeys.HASH_VAL: None,
        RemoteMMARKeys.MODEL_FILE: os.path.join("models", "model.pt"),
        RemoteMMARKeys.CONFIG_FILE: os.path.join("config", "config_train.json"),
        RemoteMMARKeys.VERSION: 1,
    },
    {
        RemoteMMARKeys.ID: "clara_pt_covid19_ct_lung_segmentation_1",
        RemoteMMARKeys.NAME: "clara_pt_covid19_ct_lung_segmentation",
        RemoteMMARKeys.FILE_TYPE: "zip",
        RemoteMMARKeys.HASH_TYPE: "md5",
        RemoteMMARKeys.HASH_VAL: None,
        RemoteMMARKeys.MODEL_FILE: os.path.join("models", "model.pt"),
        RemoteMMARKeys.CONFIG_FILE: os.path.join("config", "config_train.json"),
        RemoteMMARKeys.VERSION: 1,
    },
    {
        RemoteMMARKeys.ID: "clara_pt_unetr_ct_btcv_segmentation",
        RemoteMMARKeys.NAME: "clara_pt_unetr_ct_btcv_segmentation",
        RemoteMMARKeys.FILE_TYPE: "zip",
        RemoteMMARKeys.HASH_TYPE: "md5",
        RemoteMMARKeys.HASH_VAL: None,
        RemoteMMARKeys.MODEL_FILE: os.path.join("models", "model.pt"),
        RemoteMMARKeys.CONFIG_FILE: os.path.join("config", "config_train.json"),
        RemoteMMARKeys.VERSION: 4.1,
    },
    {
        RemoteMMARKeys.ID: "clara_pt_chest_xray_classification",
        RemoteMMARKeys.NAME: "clara_pt_chest_xray_classification",
        RemoteMMARKeys.FILE_TYPE: "zip",
        RemoteMMARKeys.HASH_TYPE: "md5",
        RemoteMMARKeys.HASH_VAL: None,
        RemoteMMARKeys.MODEL_FILE: os.path.join("models", "model.pt"),
        RemoteMMARKeys.CONFIG_FILE: os.path.join("config", "config_train.json"),
        RemoteMMARKeys.VERSION: 4.1,
    },
    {
        RemoteMMARKeys.ID: "clara_pt_self_supervised_learning_segmentation",
        RemoteMMARKeys.NAME: "clara_pt_self_supervised_learning_segmentation",
        RemoteMMARKeys.FILE_TYPE: "zip",
        RemoteMMARKeys.HASH_TYPE: "md5",
        RemoteMMARKeys.HASH_VAL: None,
        RemoteMMARKeys.MODEL_FILE: os.path.join("models_2gpu", "best_metric_model.pt"),
        RemoteMMARKeys.CONFIG_FILE: os.path.join("config", "config_train.json"),
        RemoteMMARKeys.VERSION: 4.1,
    },
)



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import json
import os
import sys
import time
from abc import ABC, abstractmethod
from collections.abc import Sequence
from copy import copy
from logging.config import fileConfig
from pathlib import Path
from typing import Any

from monai.apps.utils import get_logger
from monai.bundle.config_parser import ConfigParser
from monai.bundle.properties import InferProperties, MetaProperties, TrainProperties
from monai.bundle.utils import DEFAULT_EXP_MGMT_SETTINGS, EXPR_KEY, ID_REF_KEY, ID_SEP_KEY
from monai.config import PathLike
from monai.utils import BundleProperty, BundlePropertyConfig, deprecated_arg, ensure_tuple

__all__ = ["BundleWorkflow", "ConfigWorkflow"]

logger = get_logger(module_name=__name__)


class BundleWorkflow(ABC):
    """
    Base class for the workflow specification in bundle, it can be a training, evaluation or inference workflow.
    It defines the basic interfaces for the bundle workflow behavior: `initialize`, `run`, `finalize`, etc.
    And also provides the interface to get / set public properties to interact with a bundle workflow.

    Args:
        workflow_type: specifies the workflow type: "train" or "training" for a training workflow,
            or "infer", "inference", "eval", "evaluation" for a inference workflow,
            other unsupported string will raise a ValueError.
            default to `None` for only using meta properties.
        workflow: specifies the workflow type: "train" or "training" for a training workflow,
            or "infer", "inference", "eval", "evaluation" for a inference workflow,
            other unsupported string will raise a ValueError.
            default to `None` for common workflow.
        properties_path: the path to the JSON file of properties. If `workflow_type` is specified, properties will be
            loaded from the file based on the provided `workflow_type` and meta. If no `workflow_type` is specified,
            properties will default to loading from "meta". If `properties_path` is None, default properties
            will be sourced from "monai/bundle/properties.py" based on the workflow_type:
            For a training workflow, properties load from `TrainProperties` and `MetaProperties`.
            For a inference workflow, properties load from `InferProperties` and `MetaProperties`.
            For workflow_type = None : only `MetaProperties` will be loaded.
        meta_file: filepath of the metadata file, if this is a list of file paths, their contents will be merged in order.
        logging_file: config file for `logging` module in the program. for more details:
            https://docs.python.org/3/library/logging.config.html#logging.config.fileConfig.

    """

    supported_train_type: tuple = ("train", "training")
    supported_infer_type: tuple = ("infer", "inference", "eval", "evaluation")

    @deprecated_arg(
        "workflow",
        since="1.2",
        removed="1.5",
        new_name="workflow_type",
        msg_suffix="please use `workflow_type` instead.",
    )
    def __init__(
        self,
        workflow_type: str | None = None,
        workflow: str | None = None,
        properties_path: PathLike | None = None,
        meta_file: str | Sequence[str] | None = None,
        logging_file: str | None = None,
    ):
        if logging_file is not None:
            if not os.path.isfile(logging_file):
                raise FileNotFoundError(f"Cannot find the logging config file: {logging_file}.")
            logger.info(f"Setting logging properties based on config: {logging_file}.")
            fileConfig(logging_file, disable_existing_loggers=False)

        if meta_file is not None:
            if isinstance(meta_file, str) and not os.path.isfile(meta_file):
                logger.error(
                    f"Cannot find the metadata config file: {meta_file}. "
                    "Please see: https://docs.monai.io/en/stable/mb_specification.html"
                )
                meta_file = None
            if isinstance(meta_file, list):
                for f in meta_file:
                    if not os.path.isfile(f):
                        logger.error(
                            f"Cannot find the metadata config file: {f}. "
                            "Please see: https://docs.monai.io/en/stable/mb_specification.html"
                        )
                        meta_file = None

        workflow_type = workflow if workflow is not None else workflow_type
        if workflow_type is not None:
            if workflow_type.lower() in self.supported_train_type:
                workflow_type = "train"
            elif workflow_type.lower() in self.supported_infer_type:
                workflow_type = "infer"
            else:
                raise ValueError(f"Unsupported workflow type: '{workflow_type}'.")

        if properties_path is not None:
            properties_path = Path(properties_path)
            if not properties_path.is_file():
                raise ValueError(f"Property file {properties_path} does not exist.")
            with open(properties_path) as json_file:
                try:
                    properties = json.load(json_file)
                    self.properties: dict = {}
                    if workflow_type is not None and workflow_type in properties:
                        self.properties = properties[workflow_type]
                        if "meta" in properties:
                            self.properties.update(properties["meta"])
                    elif workflow_type is None:
                        if "meta" in properties:
                            self.properties = properties["meta"]
                            logger.info(
                                "No workflow type specified, default to load meta properties from property file."
                            )
                        else:
                            logger.warning("No 'meta' key found in properties while workflow_type is None.")
                except KeyError as e:
                    raise ValueError(f"{workflow_type} not found in property file {properties_path}") from e
                except json.JSONDecodeError as e:
                    raise ValueError(f"Error decoding JSON from property file {properties_path}") from e
        else:
            if workflow_type == "train":
                self.properties = {**TrainProperties, **MetaProperties}
            elif workflow_type == "infer":
                self.properties = {**InferProperties, **MetaProperties}
            elif workflow_type is None:
                self.properties = copy(MetaProperties)
                logger.info("No workflow type and property file specified, default to 'meta' properties.")
            else:
                raise ValueError(f"Unsupported workflow type: '{workflow_type}'.")

        self.workflow_type = workflow_type
        self.meta_file = meta_file

    @abstractmethod
    def initialize(self, *args: Any, **kwargs: Any) -> Any:
        """
        Initialize the bundle workflow before running.

        """
        raise NotImplementedError()

    @abstractmethod
    def run(self, *args: Any, **kwargs: Any) -> Any:
        """
        Run the bundle workflow, it can be a training, evaluation or inference.

        """
        raise NotImplementedError()

    @abstractmethod
    def finalize(self, *args: Any, **kwargs: Any) -> Any:
        """
        Finalize step after the running of bundle workflow.

        """
        raise NotImplementedError()

    @abstractmethod
    def _get_property(self, name: str, property: dict) -> Any:
        """
        With specified property name and information, get the expected property value.

        Args:
            name: the name of target property.
            property: other information for the target property, defined in `TrainProperties` or `InferProperties`.

        """
        raise NotImplementedError()

    @abstractmethod
    def _set_property(self, name: str, property: dict, value: Any) -> Any:
        """
        With specified property name and information, set value for the expected property.

        Args:
            name: the name of target property.
            property: other information for the target property, defined in `TrainProperties` or `InferProperties`.
            value: value to set for the property.

        """
        raise NotImplementedError()

    def __getattr__(self, name):
        if self.properties is not None and name in self.properties:
            return self._get_property(name=name, property=self.properties[name])
        else:
            return self.__getattribute__(name)  # getting regular attribute

    def __setattr__(self, name, value):
        if name != "properties" and self.properties is not None and name in self.properties:
            self._set_property(name=name, property=self.properties[name], value=value)
        else:
            super().__setattr__(name, value)  # setting regular attribute

    def get_workflow_type(self):
        """
        Get the workflow type, it can be `None`, "train", or "infer".

        """
        return self.workflow_type

    def get_meta_file(self):
        """
        Get the meta file.

        """
        return self.meta_file

    def add_property(self, name: str, required: str, desc: str | None = None) -> None:
        """
        Besides the default predefined properties, some 3rd party applications may need the bundle
        definition to provide additional properties for the specific use cases, if the bundle can't
        provide the property, means it can't work with the application.
        This utility adds the property for the application requirements check and access.

        Args:
            name: the name of target property.
            required: whether the property is "must-have".
            desc: descriptions for the property.
        """
        if self.properties is None:
            self.properties = {}
        if name in self.properties:
            logger.warn(f"property '{name}' already exists in the properties list, overriding it.")
        self.properties[name] = {BundleProperty.DESC: desc, BundleProperty.REQUIRED: required}

    def check_properties(self) -> list[str] | None:
        """
        Check whether the required properties are existing in the bundle workflow.
        If no workflow type specified, return None, otherwise, return a list of required but missing properties.

        """
        if self.properties is None:
            return None
        return [n for n, p in self.properties.items() if p.get(BundleProperty.REQUIRED, False) and not hasattr(self, n)]


class PythonicWorkflow(BundleWorkflow):
    """
    Base class for the pythonic workflow specification in bundle, it can be a training, evaluation or inference workflow.
    It defines the basic interfaces for the bundle workflow behavior: `initialize`, `finalize`, etc.
    This also provides the interface to get / set public properties to interact with a bundle workflow through
    defined `get_<property>` accessor methods or directly defining members of the object.
    For how to set the properties, users can define the `_set_<property>` methods or directly set the members of the object.
    The `initialize` method is called to set up the workflow before running. This method sets up internal state
    and prepares properties. If properties are modified after the workflow has been initialized, `self._is_initialized`
    is set to `False`. Before running the workflow again, `initialize` should be called to ensure that the workflow is
    properly set up with the new property values.

    Args:
        workflow_type: specifies the workflow type: "train" or "training" for a training workflow,
            or "infer", "inference", "eval", "evaluation" for a inference workflow,
            other unsupported string will raise a ValueError.
            default to `None` for only using meta properties.
        workflow: specifies the workflow type: "train" or "training" for a training workflow,
            or "infer", "inference", "eval", "evaluation" for a inference workflow,
            other unsupported string will raise a ValueError.
            default to `None` for common workflow.
        properties_path: the path to the JSON file of properties. If `workflow_type` is specified, properties will be
            loaded from the file based on the provided `workflow_type` and meta. If no `workflow_type` is specified,
            properties will default to loading from "meta". If `properties_path` is None, default properties
            will be sourced from "monai/bundle/properties.py" based on the workflow_type:
            For a training workflow, properties load from `TrainProperties` and `MetaProperties`.
            For a inference workflow, properties load from `InferProperties` and `MetaProperties`.
            For workflow_type = None : only `MetaProperties` will be loaded.
        config_file: path to the config file, typically used to store hyperparameters.
        meta_file: filepath of the metadata file, if this is a list of file paths, their contents will be merged in order.
        logging_file: config file for `logging` module in the program. for more details:
            https://docs.python.org/3/library/logging.config.html#logging.config.fileConfig.

    """

    supported_train_type: tuple = ("train", "training")
    supported_infer_type: tuple = ("infer", "inference", "eval", "evaluation")

    def __init__(
        self,
        workflow_type: str | None = None,
        properties_path: PathLike | None = None,
        config_file: str | Sequence[str] | None = None,
        meta_file: str | Sequence[str] | None = None,
        logging_file: str | None = None,
        **override: Any,
    ):
        meta_file = str(Path(os.getcwd()) / "metadata.json") if meta_file is None else meta_file
        super().__init__(
            workflow_type=workflow_type, properties_path=properties_path, meta_file=meta_file, logging_file=logging_file
        )
        self._props_vals: dict = {}
        self._set_props_vals: dict = {}
        self.parser = ConfigParser()
        if config_file is not None:
            self.parser.read_config(f=config_file)
        if self.meta_file is not None:
            self.parser.read_meta(f=self.meta_file)

        # the rest key-values in the _args are to override config content
        self.parser.update(pairs=override)
        self._is_initialized: bool = False

    def initialize(self, *args: Any, **kwargs: Any) -> Any:
        """
        Initialize the bundle workflow before running.
        """
        self._props_vals = {}
        self._is_initialized = True

    def _get_property(self, name: str, property: dict) -> Any:
        """
        With specified property name and information, get the expected property value.
        If the property is already generated, return from the bucket directly.
        If user explicitly set the property, return it directly.
        Otherwise, generate the expected property as a class private property with prefix "_".

        Args:
            name: the name of target property.
            property: other information for the target property, defined in `TrainProperties` or `InferProperties`.
        """
        if not self._is_initialized:
            raise RuntimeError("Please execute 'initialize' before getting any properties.")
        value = None
        if name in self._set_props_vals:
            value = self._set_props_vals[name]
        elif name in self._props_vals:
            value = self._props_vals[name]
        elif name in self.parser.config[self.parser.meta_key]:  # type: ignore[index]
            id = self.properties.get(name, None).get(BundlePropertyConfig.ID, None)
            value = self.parser[id]
        else:
            try:
                value = getattr(self, f"get_{name}")()
            except AttributeError as e:
                if property[BundleProperty.REQUIRED]:
                    raise ValueError(
                        f"unsupported property '{name}' is required in the bundle properties,"
                        f"need to implement a method 'get_{name}' to provide the property."
                    ) from e
            self._props_vals[name] = value
        return value

    def _set_property(self, name: str, property: dict, value: Any) -> Any:
        """
        With specified property name and information, set value for the expected property.
        Stores user-reset initialized objects that should not be re-initialized and marks the workflow as not initialized.

        Args:
            name: the name of target property.
            property: other information for the target property, defined in `TrainProperties` or `InferProperties`.
            value: value to set for the property.

        """
        self._set_props_vals[name] = value
        self._is_initialized = False


class ConfigWorkflow(BundleWorkflow):
    """
    Specification for the config-based bundle workflow.
    Standardized the `initialize`, `run`, `finalize` behavior in a config-based training, evaluation, or inference.
    Before `run`, we add bundle root directory to Python search directories automatically.
    For more information: https://docs.monai.io/en/latest/mb_specification.html.

    Args:
        config_file: filepath of the config file, if this is a list of file paths, their contents will be merged in order.
        meta_file: filepath of the metadata file, if this is a list of file paths, their contents will be merged in order.
            If None, default to "configs/metadata.json", which is commonly used for bundles in MONAI model zoo.
        logging_file: config file for `logging` module in the program. for more details:
            https://docs.python.org/3/library/logging.config.html#logging.config.fileConfig.
            If None, default to "configs/logging.conf", which is commonly used for bundles in MONAI model zoo.
            If False, the logging logic for the bundle will not be modified.
        init_id: ID name of the expected config expression to initialize before running, default to "initialize".
            allow a config to have no `initialize` logic and the ID.
        run_id: ID name of the expected config expression to run, default to "run".
            to run the config, the target config must contain this ID.
        final_id: ID name of the expected config expression to finalize after running, default to "finalize".
            allow a config to have no `finalize` logic and the ID.
        tracking: if not None, enable the experiment tracking at runtime with optionally configurable and extensible.
            if "mlflow", will add `MLFlowHandler` to the parsed bundle with default tracking settings,
            if other string, treat it as file path to load the tracking settings.
            if `dict`, treat it as tracking settings.
            will patch the target config content with `tracking handlers` and the top-level items of `configs`.
            for detailed usage examples, please check the tutorial:
            https://github.com/Project-MONAI/tutorials/blob/main/experiment_management/bundle_integrate_mlflow.ipynb.
        workflow_type: specifies the workflow type: "train" or "training" for a training workflow,
            or "infer", "inference", "eval", "evaluation" for a inference workflow,
            other unsupported string will raise a ValueError.
            default to `None` for common workflow.
        workflow: specifies the workflow type: "train" or "training" for a training workflow,
            or "infer", "inference", "eval", "evaluation" for a inference workflow,
            other unsupported string will raise a ValueError.
            default to `None` for common workflow.
        properties_path: the path to the JSON file of properties. If `workflow_type` is specified, properties will be
            loaded from the file based on the provided `workflow_type` and meta. If no `workflow_type` is specified,
            properties will default to loading from "train". If `properties_path` is None, default properties
            will be sourced from "monai/bundle/properties.py" based on the workflow_type:
            For a training workflow, properties load from `TrainProperties` and `MetaProperties`.
            For a inference workflow, properties load from `InferProperties` and `MetaProperties`.
            For workflow_type = None : only `MetaProperties` will be loaded.
        override: id-value pairs to override or add the corresponding config content.
            e.g. ``--net#input_chns 42``, ``--net %/data/other.json#net_arg``

    """

    @deprecated_arg(
        "workflow",
        since="1.2",
        removed="1.5",
        new_name="workflow_type",
        msg_suffix="please use `workflow_type` instead.",
    )
    def __init__(
        self,
        config_file: str | Sequence[str],
        meta_file: str | Sequence[str] | None = None,
        logging_file: str | bool | None = None,
        init_id: str = "initialize",
        run_id: str = "run",
        final_id: str = "finalize",
        tracking: str | dict | None = None,
        workflow_type: str | None = "train",
        workflow: str | None = None,
        properties_path: PathLike | None = None,
        **override: Any,
    ) -> None:
        workflow_type = workflow if workflow is not None else workflow_type
        if config_file is not None:
            _config_files = ensure_tuple(config_file)
            config_root_path = Path(_config_files[0]).parent
            for _config_file in _config_files:
                _config_file = Path(_config_file)
                if _config_file.parent != config_root_path:
                    logger.warn(
                        f"Not all config files are in {config_root_path}. If logging_file and meta_file are"
                        f"not specified, {config_root_path} will be used as the default config root directory."
                    )
                if not _config_file.is_file():
                    raise FileNotFoundError(f"Cannot find the config file: {_config_file}.")
        else:
            config_root_path = Path("configs")
        meta_file = str(config_root_path / "metadata.json") if meta_file is None else meta_file
        super().__init__(workflow_type=workflow_type, meta_file=meta_file, properties_path=properties_path)
        self.config_root_path = config_root_path
        logging_file = str(self.config_root_path / "logging.conf") if logging_file is None else logging_file
        if logging_file is False:
            logger.warn(f"Logging file is set to {logging_file}, skipping logging.")
        else:
            if not os.path.isfile(logging_file):
                if logging_file == str(self.config_root_path / "logging.conf"):
                    logger.warn(f"Default logging file in {logging_file} does not exist, skipping logging.")
                else:
                    raise FileNotFoundError(f"Cannot find the logging config file: {logging_file}.")
            else:
                fileConfig(str(logging_file), disable_existing_loggers=False)
                logger.info(f"Setting logging properties based on config: {logging_file}.")

        self.parser = ConfigParser()
        self.parser.read_config(f=config_file)
        if self.meta_file is not None:
            self.parser.read_meta(f=self.meta_file)
        # the rest key-values in the _args are to override config content
        self.parser.update(pairs=override)
        self.init_id = init_id
        self.run_id = run_id
        self.final_id = final_id
        # set tracking configs for experiment management
        if tracking is not None:
            if isinstance(tracking, str) and tracking in DEFAULT_EXP_MGMT_SETTINGS:
                settings_ = DEFAULT_EXP_MGMT_SETTINGS[tracking]
            else:
                settings_ = ConfigParser.load_config_files(tracking)
            self.patch_bundle_tracking(parser=self.parser, settings=settings_)
        self._is_initialized: bool = False

    def initialize(self) -> Any:
        """
        Initialize the bundle workflow before running.

        """
        # reset the "reference_resolver" buffer at initialization stage
        self.parser.parse(reset=True)
        self._is_initialized = True
        return self._run_expr(id=self.init_id)

    def run(self) -> Any:
        """
        Run the bundle workflow, it can be a training, evaluation or inference.
        Before run, we add bundle root directory to Python search directories automatically.

        """
        _bundle_root_path = (
            self.config_root_path.parent if self.config_root_path.name == "configs" else self.config_root_path
        )
        sys.path.insert(1, str(_bundle_root_path))
        if self.run_id not in self.parser:
            raise ValueError(f"run ID '{self.run_id}' doesn't exist in the config file.")
        return self._run_expr(id=self.run_id)

    def finalize(self) -> Any:
        """
        Finalize step after the running of bundle workflow.

        """
        return self._run_expr(id=self.final_id)

    def check_properties(self) -> list[str] | None:
        """
        Check whether the required properties are existing in the bundle workflow.
        If the optional properties have reference in the config, will also check whether the properties are existing.
        If no workflow type specified, return None, otherwise, return a list of required but missing properties.

        """
        ret = super().check_properties()
        if self.properties is None:
            logger.warn("No available properties had been set, skipping check.")
            return None
        if ret:
            logger.warn(f"Loaded bundle does not contain the following required properties: {ret}")
        # also check whether the optional properties use correct ID name if existing
        wrong_props = []
        for n, p in self.properties.items():
            if not p.get(BundleProperty.REQUIRED, False) and not self._check_optional_id(name=n, property=p):
                wrong_props.append(n)
        if wrong_props:
            logger.warn(f"Loaded bundle defines the following optional properties with wrong ID: {wrong_props}")
        if ret is not None:
            ret.extend(wrong_props)
        return ret

    def _run_expr(self, id: str, **kwargs: dict) -> list[Any]:
        """
        Evaluate the expression or expression list given by `id`. The resolved values from the evaluations are not stored,
        allowing this to be evaluated repeatedly (eg. in streaming applications) without restarting the hosting process.
        """
        ret = []
        if id in self.parser:
            # suppose all the expressions are in a list, run and reset the expressions
            if isinstance(self.parser[id], list):
                for i in range(len(self.parser[id])):
                    sub_id = f"{id}{ID_SEP_KEY}{i}"
                    ret.append(self.parser.get_parsed_content(sub_id, **kwargs))
                    self.parser.ref_resolver.remove_resolved_content(sub_id)
            else:
                ret.append(self.parser.get_parsed_content(id, **kwargs))
                self.parser.ref_resolver.remove_resolved_content(id)
        return ret

    def _get_prop_id(self, name: str, property: dict) -> Any:
        prop_id = property[BundlePropertyConfig.ID]
        if prop_id not in self.parser:
            if not property.get(BundleProperty.REQUIRED, False):
                return None
            else:
                raise KeyError(f"Property '{name}' with config ID '{prop_id}' not in the config.")
        return prop_id

    def _get_property(self, name: str, property: dict) -> Any:
        """
        With specified property name and information, get the parsed property value from config.

        Args:
            name: the name of target property.
            property: other information for the target property, defined in `TrainProperties` or `InferProperties`.

        """
        if not self._is_initialized:
            raise RuntimeError("Please execute 'initialize' before getting any parsed content.")
        prop_id = self._get_prop_id(name, property)
        return self.parser.get_parsed_content(id=prop_id) if prop_id is not None else None

    def _set_property(self, name: str, property: dict, value: Any) -> None:
        """
        With specified property name and information, set value for the expected property.

        Args:
            name: the name of target property.
            property: other information for the target property, defined in `TrainProperties` or `InferProperties`.
            value: value to set for the property.

        """
        prop_id = self._get_prop_id(name, property)
        if prop_id is not None:
            self.parser[prop_id] = value
            # must parse the config again after changing the content
            self._is_initialized = False
            self.parser.ref_resolver.reset()

    def add_property(  # type: ignore[override]
        self, name: str, required: str, config_id: str, desc: str | None = None
    ) -> None:
        """
        Besides the default predefined properties, some 3rd party applications may need the bundle
        definition to provide additional properties for the specific use cases, if the bundle can't
        provide the property, means it can't work with the application.
        This utility adds the property for the application requirements check and access.

        Args:
            name: the name of target property.
            required: whether the property is "must-have".
            config_id: the config ID of target property in the bundle definition.
            desc: descriptions for the property.

        """
        super().add_property(name=name, required=required, desc=desc)
        self.properties[name][BundlePropertyConfig.ID] = config_id

    def _check_optional_id(self, name: str, property: dict) -> bool:
        """
        If an optional property has reference in the config, check whether the property is existing.
        If `ValidationHandler` is defined for a training workflow, will check whether the optional properties
        "evaluator" and "val_interval" are existing.

        Args:
            name: the name of target property.
            property: other information for the target property, defined in `TrainProperties` or `InferProperties`.

        """
        id = property.get(BundlePropertyConfig.ID, None)
        ref_id = property.get(BundlePropertyConfig.REF_ID, None)
        if ref_id is None:
            # no ID of reference config item, skipping check for this optional property
            return True
        # check validation `validator` and `interval` properties as the handler index of ValidationHandler is unknown
        ref: str | None = None
        if name in ("evaluator", "val_interval"):
            if f"train{ID_SEP_KEY}handlers" in self.parser:
                for h in self.parser[f"train{ID_SEP_KEY}handlers"]:
                    if h["_target_"] == "ValidationHandler":
                        ref = h.get(ref_id, None)
        else:
            ref = self.parser.get(ref_id, None)
        # for reference IDs that not refer to a property directly but using expressions, skip the check
        if ref is not None and not ref.startswith(EXPR_KEY) and ref != ID_REF_KEY + id:
            return False
        return True

    @staticmethod
    def patch_bundle_tracking(parser: ConfigParser, settings: dict) -> None:
        """
        Patch the loaded bundle config with a new handler logic to enable experiment tracking features.

        Args:
            parser: loaded config content to patch the handler.
            settings: settings for the experiment tracking, should follow the pattern of default settings.

        """
        for k, v in settings["configs"].items():
            if k in settings["handlers_id"]:
                engine = parser.get(settings["handlers_id"][k]["id"])
                if engine is not None:
                    handlers = parser.get(settings["handlers_id"][k]["handlers"])
                    if handlers is None:
                        engine["train_handlers" if k == "trainer" else "val_handlers"] = [v]
                    else:
                        handlers.append(v)
            elif k not in parser:
                parser[k] = v
        # save the executed config into file
        default_name = f"config_{time.strftime('%Y%m%d_%H%M%S')}.json"
        # Users can set the `save_execute_config` to `False`, `/path/to/artifacts` or `True`.
        # If set to False, nothing will be recorded. If set to True, the default path will be logged.
        # If set to a file path, the given path will be logged.
        filepath = parser.get("save_execute_config", True)
        if filepath:
            if isinstance(filepath, bool):
                if "output_dir" not in parser:
                    # if no "output_dir" in the bundle config, default to "<bundle root>/eval"
                    parser["output_dir"] = f"{EXPR_KEY}{ID_REF_KEY}bundle_root + '/eval'"
                # experiment management tools can refer to this config item to track the config info
                parser["save_execute_config"] = parser["output_dir"] + f" + '/{default_name}'"
                filepath = os.path.join(parser.get_parsed_content("output_dir"), default_name)
            Path(filepath).parent.mkdir(parents=True, exist_ok=True)
            parser.export_config_file(parser.get(), filepath)
        else:
            parser["save_execute_config"] = None



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import tempfile
import unittest

import torch
from parameterized import parameterized

from monai.networks import convert_to_trt
from monai.networks.nets import UNet
from monai.utils import optional_import
from tests.test_utils import SkipIfBeforeComputeCapabilityVersion, skip_if_no_cuda, skip_if_quick, skip_if_windows

_, has_torchtrt = optional_import(
    "torch_tensorrt",
    version="1.4.0",
    descriptor="Torch-TRT is not installed. Are you sure you have a Torch-TensorRT compilation?",
)
_, has_tensorrt = optional_import(
    "tensorrt", descriptor="TensorRT is not installed. Are you sure you have a TensorRT compilation?"
)

TEST_CASE_1 = ["fp32"]
TEST_CASE_2 = ["fp16"]


@skip_if_windows
@skip_if_no_cuda
@skip_if_quick
@SkipIfBeforeComputeCapabilityVersion((7, 5))
class TestConvertToTRT(unittest.TestCase):
    def setUp(self):
        self.gpu_device = torch.cuda.current_device()

    def tearDown(self):
        current_device = torch.cuda.current_device()
        if current_device != self.gpu_device:
            torch.cuda.set_device(self.gpu_device)

    @parameterized.expand([TEST_CASE_1, TEST_CASE_2])
    @unittest.skipUnless(has_torchtrt and has_tensorrt, "Torch-TensorRT is required for convert!")
    def test_value(self, precision):
        model = UNet(
            spatial_dims=3,
            in_channels=1,
            out_channels=2,
            channels=(2, 2, 4, 8, 4),
            strides=(2, 2, 2, 2),
            num_res_units=2,
            norm="batch",
        )
        with tempfile.TemporaryDirectory() as _:
            torchscript_model = convert_to_trt(
                model=model,
                precision=precision,
                input_shape=[1, 1, 96, 96, 96],
                dynamic_batchsize=[1, 4, 8],
                use_trace=False,
                verify=True,
                device=0,
                rtol=1e-2,
                atol=1e-2,
            )
            self.assertTrue(isinstance(torchscript_model, torch.nn.Module))


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from collections.abc import Sequence

import numpy as np
import torch
import torch.nn as nn

from monai.networks.blocks import Convolution, ResidualUnit
from monai.networks.layers.factories import Act, Norm
from monai.networks.layers.simplelayers import Reshape
from monai.utils import ensure_tuple, ensure_tuple_rep


class Generator(nn.Module):
    """
    Defines a simple generator network accepting a latent vector and through a sequence of convolution layers
    constructs an output tensor of greater size and high dimensionality. The method `_get_layer` is used to
    create each of these layers, override this method to define layers beyond the default
    :py:class:`monai.networks.blocks.Convolution` or :py:class:`monai.networks.blocks.ResidualUnit` layers.

    The layers are constructed using the values in the `channels` and `strides` arguments, the number being defined by
    the length of these (which must match). Input is first passed through a :py:class:`torch.nn.Linear` layer to
    convert the input vector to an image tensor with dimensions `start_shape`. This passes through the convolution
    layers and is progressively upsampled if the `strides` values are greater than 1 using transpose convolutions. The
    size of the final output is defined by the `start_shape` dimension and the amount of upsampling done through
    strides. In the default definition the size of the output's spatial dimensions will be that of `start_shape`
    multiplied by the product of `strides`, thus the example network below upsamples an starting size of (64, 8, 8)
    to (1, 64, 64) since its `strides` are (2, 2, 2).

    Args:
        latent_shape: tuple of integers stating the dimension of the input latent vector (minus batch dimension)
        start_shape: tuple of integers stating the dimension of the tensor to pass to convolution subnetwork
        channels: tuple of integers stating the output channels of each convolutional layer
        strides: tuple of integers stating the stride (upscale factor) of each convolutional layer
        kernel_size: integer or tuple of integers stating size of convolutional kernels
        num_res_units: integer stating number of convolutions in residual units, 0 means no residual units
        act: name or type defining activation layers
        norm: name or type defining normalization layers
        dropout: optional float value in range [0, 1] stating dropout probability for layers, None for no dropout
        bias: boolean stating if convolution layers should have a bias component

    Examples::

        # 3 layers, latent input vector of shape (42, 24), output volume of shape (1, 64, 64)
        net = Generator((42, 24), (64, 8, 8), (32, 16, 1), (2, 2, 2))

    """

    def __init__(
        self,
        latent_shape: Sequence[int],
        start_shape: Sequence[int],
        channels: Sequence[int],
        strides: Sequence[int],
        kernel_size: Sequence[int] | int = 3,
        num_res_units: int = 2,
        act=Act.PRELU,
        norm=Norm.INSTANCE,
        dropout: float | None = None,
        bias: bool = True,
    ) -> None:
        super().__init__()

        self.in_channels, *self.start_shape = ensure_tuple(start_shape)
        self.dimensions = len(self.start_shape)

        self.latent_shape = ensure_tuple(latent_shape)
        self.channels = ensure_tuple(channels)
        self.strides = ensure_tuple(strides)
        self.kernel_size = ensure_tuple_rep(kernel_size, self.dimensions)
        self.num_res_units = num_res_units
        self.act = act
        self.norm = norm
        self.dropout = dropout
        self.bias = bias

        self.flatten = nn.Flatten()
        self.linear = nn.Linear(int(np.prod(self.latent_shape)), int(np.prod(start_shape)))
        self.reshape = Reshape(*start_shape)
        self.conv = nn.Sequential()

        echannel = self.in_channels

        # transform tensor of shape `start_shape' into output shape through transposed convolutions and residual units
        for i, (c, s) in enumerate(zip(channels, strides)):
            is_last = i == len(channels) - 1
            layer = self._get_layer(echannel, c, s, is_last)
            self.conv.add_module("layer_%i" % i, layer)
            echannel = c

    def _get_layer(
        self, in_channels: int, out_channels: int, strides: int, is_last: bool
    ) -> Convolution | nn.Sequential:
        """
        Returns a layer accepting inputs with `in_channels` number of channels and producing outputs of `out_channels`
        number of channels. The `strides` indicates upsampling factor, ie. transpose convolutional stride. If `is_last`
        is True this is the final layer and is not expected to include activation and normalization layers.
        """

        layer: Convolution | nn.Sequential

        layer = Convolution(
            in_channels=in_channels,
            strides=strides,
            is_transposed=True,
            conv_only=is_last or self.num_res_units > 0,
            spatial_dims=self.dimensions,
            out_channels=out_channels,
            kernel_size=self.kernel_size,
            act=self.act,
            norm=self.norm,
            dropout=self.dropout,
            bias=self.bias,
        )

        if self.num_res_units > 0:
            ru = ResidualUnit(
                in_channels=out_channels,
                subunits=self.num_res_units,
                last_conv_only=is_last,
                spatial_dims=self.dimensions,
                out_channels=out_channels,
                kernel_size=self.kernel_size,
                act=self.act,
                norm=self.norm,
                dropout=self.dropout,
                bias=self.bias,
            )

            layer = nn.Sequential(layer, ru)

        return layer

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.flatten(x)
        x = self.linear(x)
        x = self.reshape(x)
        x = self.conv(x)
        return x



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import unittest
from typing import Any

import numpy as np
import torch
from parameterized import parameterized

from monai.metrics import (
    ConfusionMatrixMetric,
    compute_confusion_matrix_metric,
    do_metric_reduction,
    get_confusion_matrix,
)
from tests.test_utils import assert_allclose

_device = "cuda:0" if torch.cuda.is_available() else "cpu"
# input data
data: dict[Any, Any] = {
    "y_pred": torch.tensor(
        [
            [[[0.0, 1.0], [0.0, 0.0]], [[0.0, 0.0], [1.0, 1.0]], [[1.0, 0.0], [0.0, 0.0]]],
            [[[0.0, 0.0], [0.0, 1.0]], [[1.0, 0.0], [0.0, 0.0]], [[0.0, 1.0], [1.0, 0.0]]],
        ],
        device=_device,
    ),
    "y": torch.tensor(
        [
            [[[0.0, 0.0], [0.0, 1.0]], [[1.0, 0.0], [1.0, 0.0]], [[0.0, 1.0], [0.0, 0.0]]],
            [[[0.0, 0.0], [0.0, 1.0]], [[1.0, 1.0], [0.0, 0.0]], [[0.0, 0.0], [1.0, 0.0]]],
        ],
        device=_device,
    ),
}

data_nan: dict[Any, Any] = {
    # confusion matrix:[[[0,1,2,1],[1,1,1,1],[0,1,2,1]],
    #                   [[0,0,0,4],[0,0,4,0],[0,4,0,0]],
    #                   [[0,0,2,2],[0,0,2,2],[0,4,0,0]]]
    "y_pred": torch.tensor(
        [
            [[[0.0, 1.0], [0.0, 0.0]], [[0.0, 0.0], [1.0, 1.0]], [[1.0, 0.0], [0.0, 0.0]]],
            [[[0.0, 0.0], [0.0, 0.0]], [[0.0, 0.0], [0.0, 0.0]], [[1.0, 1.0], [1.0, 1.0]]],
            [[[0.0, 0.0], [0.0, 0.0]], [[0.0, 0.0], [0.0, 0.0]], [[1.0, 1.0], [1.0, 1.0]]],
        ]
    ),
    "y": torch.tensor(
        [
            [[[0.0, 0.0], [0.0, 1.0]], [[1.0, 0.0], [1.0, 0.0]], [[0.0, 1.0], [0.0, 0.0]]],
            [[[1.0, 1.0], [1.0, 1.0]], [[0.0, 0.0], [0.0, 0.0]], [[0.0, 0.0], [0.0, 0.0]]],
            [[[0.0, 1.0], [1.0, 0.0]], [[1.0, 0.0], [0.0, 1.0]], [[0.0, 0.0], [0.0, 0.0]]],
        ]
    ),
}

data_clf: dict[Any, Any] = {
    "y_pred": torch.tensor([[1, 0, 0], [0, 0, 1]]),
    "y": torch.tensor([[1, 0, 0], [0, 1, 0]]),
    "compute_sample": False,
    "include_background": True,
    "metric_name": "tpr",
    "reduction": "mean_channel",
    "get_not_nans": True,
}

# 1. test confusion matrix
TEST_CASE_CONFUSION_MATRIX = [
    data.copy(),
    torch.tensor(
        [
            [[0.0, 1.0, 2.0, 1.0], [1.0, 1.0, 1.0, 1.0], [0.0, 1.0, 2.0, 1.0]],
            [[1.0, 0.0, 3.0, 0.0], [1.0, 0.0, 2.0, 1.0], [1.0, 1.0, 2.0, 0.0]],
        ]
    ),
]

# 2. test metric with compute_sample
TEST_CASES_COMPUTE_SAMPLE = []
TEST_CASES_COMPUTE_SAMPLE_MULTI_METRICS = []
result_mean_batch = [
    torch.tensor([0.5000, 0.5000, 0.5000]),
    torch.tensor([0.1667, 0.2500, 0.3333]),
    torch.tensor([0.8333, 0.7500, 0.6667]),
    torch.tensor([0.5000, 0.7500, 0.2500]),
    torch.tensor([0.8333, 0.5833, 0.8333]),
    torch.tensor([0.5000, 0.5000, 0.5000]),
    torch.tensor([0.5000, 0.2500, 0.7500]),
    torch.tensor([0.1667, 0.4167, 0.1667]),
    torch.tensor([0.5000, 0.0000, 0.6830]),
    torch.tensor([0.5000, 0.4167, 0.2500]),
    torch.tensor([0.7500, 0.6250, 0.6250]),
    torch.tensor([0.6667, 0.6250, 0.5833]),
    torch.tensor([0.5000, 0.5833, 0.3333]),
    torch.tensor([0.3333, 0.2887, 0.1220]),
    torch.tensor([0.5000, 0.6036, 0.3536]),
    torch.tensor([0.3333, 0.2500, 0.1667]),
    torch.tensor([0.3333, 0.3333, 0.0833]),
]
result_mean = [
    torch.tensor([0.5000]),
    torch.tensor([0.2500]),
    torch.tensor([0.7500]),
    torch.tensor([0.5000]),
    torch.tensor([0.7500]),
    torch.tensor([0.5000]),
    torch.tensor([0.5000]),
    torch.tensor([0.2500]),
    torch.tensor([0.5610]),
    torch.tensor([0.3889]),
    torch.tensor([0.6667]),
    torch.tensor([0.6250]),
    torch.tensor([0.4722]),
    torch.tensor([0.2480]),
    torch.tensor([0.4857]),
    torch.tensor([0.2500]),
    torch.tensor([0.2500]),
]
metric_names = [
    "tpr",
    "fpr",
    "tnr",
    "ppv",
    "npv",
    "fnr",
    "fdr",
    "for",
    "pt",
    "ts",
    "acc",
    "ba",
    "f1",
    "mcc",
    "fm",
    "bm",
    "mk",
]
result: Any = None
for idx, item in enumerate(metric_names):
    for reduction in ["mean", "mean_batch"]:
        TEST_CASE: list[Any] = [data.copy()]
        TEST_CASE[0]["compute_sample"] = True
        TEST_CASE[0]["include_background"] = True
        TEST_CASE[0]["metric_name"] = item
        TEST_CASE[0]["reduction"] = reduction
        TEST_CASE[0]["get_not_nans"] = True
        if reduction == "mean_batch":
            result = result_mean_batch[idx]
        elif reduction == "mean":
            result = result_mean[idx]
        TEST_CASE.append(result)
        TEST_CASES_COMPUTE_SAMPLE.append(TEST_CASE)

# one input to compute multiple metrics
for reduction in ["mean", "mean_batch"]:
    TEST_CASE_MULTIPLE: list[Any] = [data.copy()]
    TEST_CASE_MULTIPLE[0]["compute_sample"] = True
    TEST_CASE_MULTIPLE[0]["include_background"] = True
    TEST_CASE_MULTIPLE[0]["metric_name"] = metric_names
    TEST_CASE_MULTIPLE[0]["reduction"] = reduction
    TEST_CASE_MULTIPLE[0]["get_not_nans"] = True
    if reduction == "mean_batch":
        result = result_mean_batch
    elif reduction == "mean":
        result = result_mean
    TEST_CASE_MULTIPLE.append(result)
    TEST_CASES_COMPUTE_SAMPLE_MULTI_METRICS.append(TEST_CASE_MULTIPLE)

# 3. test metric with compute_sample, denominator may have zeros
TEST_CASES_COMPUTE_SAMPLE_NAN = []
metric_names = ["tpr", "tnr"]
result_sum = [torch.tensor([0.5000]), torch.tensor([4.8333])]
not_nans_sum = [torch.tensor([6]), torch.tensor([8])]
result_sum_batch = [torch.tensor([0.0000, 0.5000, 0.0000]), torch.tensor([1.6667, 2.5000, 0.6667])]
not_nans_sum_batch = [torch.tensor([3.0, 2.0, 1.0]), torch.tensor([2.0, 3.0, 3.0])]
for idx in range(2):
    for reduction in ["sum", "sum_batch"]:
        TEST_CASE = [data_nan.copy()]
        TEST_CASE[0]["compute_sample"] = True
        TEST_CASE[0]["include_background"] = True
        TEST_CASE[0]["reduction"] = reduction
        TEST_CASE[0]["metric_name"] = metric_names[idx]
        TEST_CASE[0]["get_not_nans"] = True
        if reduction == "sum":
            TEST_CASE.append(result_sum[idx])
            TEST_CASE.append(not_nans_sum[idx])
        elif reduction == "sum_batch":
            TEST_CASE.append(result_sum_batch[idx])
            TEST_CASE.append(not_nans_sum_batch[idx])
        TEST_CASES_COMPUTE_SAMPLE_NAN.append(TEST_CASE)

# 4. test classification task
result_clf = torch.tensor(
    [
        [[1.0, 0.0, 0.0, 0.0], [0.0, 0.0, 1.0, 0.0], [0.0, 0.0, 1.0, 0.0]],
        [[0.0, 0.0, 1.0, 0.0], [0.0, 0.0, 0.0, 1.0], [0.0, 1.0, 0.0, 0.0]],
    ]
)

TEST_CASES_CLF = [data_clf.copy(), result_clf]

TEST_CASE_PRECISION = [
    {
        "y_pred": torch.zeros([1, 1, 1024, 1024, 44], device=_device),
        "y": torch.zeros([1, 1, 1024, 1024, 44], device=_device),
    },
    torch.tensor([[[0.0, 0.0, 46137344.0, 0.0]]]),
]


class TestConfusionMatrix(unittest.TestCase):
    @parameterized.expand([TEST_CASE_CONFUSION_MATRIX])
    def test_value(self, input_data, expected_value):
        # include or ignore background
        input_data["include_background"] = True
        result = get_confusion_matrix(**input_data)
        assert_allclose(result, expected_value, atol=1e-4, rtol=1e-4)
        input_data["include_background"] = False
        result = get_confusion_matrix(**input_data)
        assert_allclose(result, expected_value[:, 1:, :], atol=1e-4, rtol=1e-4)
        np.testing.assert_equal(result.device, input_data["y_pred"].device)

    @parameterized.expand(TEST_CASES_COMPUTE_SAMPLE)
    def test_compute_sample(self, input_data, expected_value):
        params = input_data.copy()
        vals = {}
        vals["y_pred"] = params.pop("y_pred")
        vals["y"] = params.pop("y")
        metric = ConfusionMatrixMetric(**params)
        metric(**vals)
        result, _ = metric.aggregate()[0]
        assert_allclose(result, expected_value, atol=1e-4, rtol=1e-4)

    @parameterized.expand(TEST_CASES_COMPUTE_SAMPLE_MULTI_METRICS)
    def test_compute_sample_multiple_metrics(self, input_data, expected_values):
        params = input_data.copy()
        vals = {}
        vals["y_pred"] = params.pop("y_pred")
        vals["y"] = params.pop("y")
        metric = ConfusionMatrixMetric(**params)
        metric(**vals)
        results = metric.aggregate()
        for idx, item in enumerate(results):
            result = item[0]
            expected_value = expected_values[idx]
            assert_allclose(result, expected_value, atol=1e-4, rtol=1e-4)

    @parameterized.expand(TEST_CASES_COMPUTE_SAMPLE_NAN)
    def test_compute_sample_with_nan(self, input_data, expected_value, expected_not_nans):
        params = input_data.copy()
        vals = {}
        vals["y_pred"] = params.pop("y_pred")
        vals["y"] = params.pop("y")
        metric = ConfusionMatrixMetric(**params)
        metric(**vals)
        result, not_nans = metric.aggregate()[0]
        assert_allclose(result, expected_value, atol=1e-4, rtol=1e-4)
        assert_allclose(not_nans, expected_not_nans, atol=1e-4, rtol=1e-4)

    @parameterized.expand([TEST_CASES_CLF])
    def test_clf_with_nan(self, input_data, expected_value):
        params = input_data.copy()
        vals = {}
        vals["y_pred"] = params.pop("y_pred")
        vals["y"] = params.pop("y")
        metric = ConfusionMatrixMetric(**params)
        result = metric(**vals)
        assert_allclose(result, expected_value, atol=1e-4, rtol=1e-4)
        result, _ = metric.aggregate(reduction="mean_channel")[0]
        expected_value, _ = do_metric_reduction(expected_value, "mean_channel")
        expected_value = compute_confusion_matrix_metric("tpr", expected_value)
        assert_allclose(result, expected_value, atol=1e-4, rtol=1e-4)

    @parameterized.expand([TEST_CASE_PRECISION])
    def test_precision(self, input_data, expected_value):
        # include or ignore background
        result = get_confusion_matrix(**input_data)
        assert_allclose(result, expected_value, atol=1e-4, rtol=1e-4)
        np.testing.assert_equal(result.device, input_data["y_pred"].device)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import os
import shutil
import tempfile
import unittest

from monai.transforms import (
    Compose,
    CopyItemsd,
    EnsureChannelFirstd,
    Invertd,
    Lambda,
    LoadImaged,
    ResampleToMatchd,
    SaveImaged,
)
from tests.lazy_transforms_utils import test_resampler_lazy
from tests.test_utils import assert_allclose, download_url_or_skip_test, testing_data_config


def update_fname(d):
    d["im3"].meta["filename_or_obj"] = "file3.nii.gz"
    return d


class TestResampleToMatchd(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        super(__class__, cls).setUpClass()
        cls.fnames = []
        cls.tmpdir = tempfile.mkdtemp()
        for key in ("0000_t2_tse_tra_4", "0000_ep2d_diff_tra_7"):
            fname = os.path.join(cls.tmpdir, f"test_{key}.nii.gz")
            url = testing_data_config("images", key, "url")
            hash_type = testing_data_config("images", key, "hash_type")
            hash_val = testing_data_config("images", key, "hash_val")
            download_url_or_skip_test(url=url, filepath=fname, hash_type=hash_type, hash_val=hash_val)
            cls.fnames.append(fname)

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.tmpdir)
        super(__class__, cls).tearDownClass()

    def test_correct(self):
        transforms = Compose(
            [
                LoadImaged(("im1", "im2")),
                EnsureChannelFirstd(("im1", "im2")),
                CopyItemsd(("im2"), names=("im3")),
                ResampleToMatchd("im3", "im1"),
                Lambda(update_fname),
                SaveImaged("im3", output_dir=self.tmpdir, output_postfix="", separate_folder=False, resample=False),
            ]
        )
        data = transforms({"im1": self.fnames[0], "im2": self.fnames[1]})
        # check that output sizes match
        assert_allclose(data["im1"].shape, data["im3"].shape)
        # and that the meta data has been updated accordingly
        assert_allclose(data["im3"].affine, data["im1"].affine)
        # check we're different from the original
        self.assertTrue(any(i != j for i, j in zip(data["im3"].shape, data["im2"].shape)))
        self.assertTrue(any(i != j for i, j in zip(data["im3"].affine.flatten(), data["im2"].affine.flatten())))
        # test the inverse
        data = Invertd("im3", transforms)(data)
        assert_allclose(data["im2"].shape, data["im3"].shape)

    def test_lazy(self):
        pre_transforms = Compose(
            [LoadImaged(("im1", "im2")), EnsureChannelFirstd(("im1", "im2")), CopyItemsd(("im2"), names=("im3"))]
        )
        data = pre_transforms({"im1": self.fnames[0], "im2": self.fnames[1]})
        init_param = {"keys": "im3", "key_dst": "im1"}
        resampler = ResampleToMatchd(**init_param)
        call_param = {"data": data}
        non_lazy_out = resampler(**call_param)
        test_resampler_lazy(resampler, non_lazy_out, init_param, call_param, output_key="im3")


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from collections.abc import Sequence

import torch

from monai.metrics.utils import do_metric_reduction, ignore_background
from monai.utils import MetricReduction

from .metric import CumulativeIterationMetric


class FBetaScore(CumulativeIterationMetric):

    def __init__(
        self,
        beta: float = 1.0,
        include_background: bool = True,
        reduction: MetricReduction | str = MetricReduction.MEAN,
        get_not_nans: bool = False,
    ) -> None:
        super().__init__()
        self.beta = beta
        self.include_background = include_background
        self.reduction = reduction
        self.get_not_nans = get_not_nans

    def _compute_tensor(self, y_pred: torch.Tensor, y: torch.Tensor) -> torch.Tensor:  # type: ignore[override]
        if y_pred.ndimension() < 2:
            raise ValueError("y_pred should have at least two dimensions.")

        return get_f_beta_score(y_pred=y_pred, y=y, include_background=self.include_background)

    def aggregate(
        self, compute_sample: bool = False, reduction: MetricReduction | str | None = None
    ) -> Sequence[torch.Tensor | tuple[torch.Tensor, torch.Tensor]]:
        data = self.get_buffer()
        if not isinstance(data, torch.Tensor):
            raise ValueError("the data to aggregate must be PyTorch Tensor.")

        results: list[torch.Tensor | tuple[torch.Tensor, torch.Tensor]] = []
        f, not_nans = do_metric_reduction(data, reduction or self.reduction)
        f = compute_f_beta_score(f, self.beta)
        if self.get_not_nans:
            results.append((f, not_nans))
        else:
            results.append(f)

        return results


def get_f_beta_score(y_pred: torch.Tensor, y: torch.Tensor, include_background: bool = True) -> torch.Tensor:
    if not include_background:
        y_pred, y = ignore_background(y_pred=y_pred, y=y)

    if y.shape != y_pred.shape:
        raise ValueError(f"y_pred and y should have same shapes, got {y_pred.shape} and {y.shape}.")

    # get confusion matrix related metric
    batch_size, n_class = y_pred.shape[:2]
    # convert to [BNS], where S is the number of pixels for one sample.
    # As for classification tasks, S equals to 1.
    y_pred = y_pred.view(batch_size, n_class, -1)
    y = y.view(batch_size, n_class, -1)
    tp = (y_pred + y) == 2
    tn = (y_pred + y) == 0

    tp = tp.sum(dim=[2]).float()
    tn = tn.sum(dim=[2]).float()
    p = y.sum(dim=[2]).float()
    n = y.shape[-1] - p

    fn = p - tp
    fp = n - tn

    return torch.stack([tp, fp, tn, fn], dim=-1)


def compute_f_beta_score(confusion_matrix: torch.Tensor, beta: float) -> torch.Tensor:
    input_dim = confusion_matrix.ndimension()
    if input_dim == 1:
        confusion_matrix = confusion_matrix.unsqueeze(dim=0)
    if confusion_matrix.shape[-1] != 4:
        raise ValueError("the size of the last dimension of confusion_matrix should be 4.")

    tp = confusion_matrix[..., 0]
    fp = confusion_matrix[..., 1]
    # tn = confusion_matrix[..., 2]
    fn = confusion_matrix[..., 3]

    nan_tensor = torch.tensor(float("nan"), device=confusion_matrix.device)
    numerator, denominator = (1.0 + beta**2) * tp, ((1.0 + beta**2) * tp + beta**2 * fn + fp)

    if isinstance(denominator, torch.Tensor):
        return torch.where(denominator != 0, numerator / denominator, nan_tensor)
    return numerator / denominator



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import sys
import time
import unittest

from parameterized import parameterized

from monai.transforms import DeleteItemsd
from monai.utils.enums import PostFix

TEST_CASE_1 = [{"keys": [str(i) for i in range(30)]}, 20]

TEST_CASE_2 = [{"keys": ["image/" + str(i) for i in range(30)], "sep": "/"}, 20]

TEST_CASE_3 = [{"keys": "meta_dict%0008\\|[0-9]", "sep": "%", "use_re": True}]


class TestDeleteItemsd(unittest.TestCase):

    @parameterized.expand([TEST_CASE_1, TEST_CASE_2])
    def test_memory(self, input_param, expected_key_size):
        input_data = {"image": {}} if "sep" in input_param else {}
        for i in range(50):
            if "sep" in input_param:
                input_data["image"][str(i)] = [time.time()] * 100000
            else:
                input_data[str(i)] = [time.time()] * 100000
        result = DeleteItemsd(**input_param)(input_data)
        if "sep" in input_param:
            self.assertEqual(len(result["image"].keys()), expected_key_size)
        else:
            self.assertEqual(len(result.keys()), expected_key_size)
        self.assertGreaterEqual(
            sys.getsizeof(input_data) * float(expected_key_size) / len(input_data), sys.getsizeof(result)
        )

    @parameterized.expand([TEST_CASE_3])
    def test_re(self, input_param):
        input_data = {"image": [1, 2, 3], PostFix.meta(): {"0008|0005": 1, "0008|1050": 2, "0008test": 3}}
        result = DeleteItemsd(**input_param)(input_data)
        self.assertEqual(result[PostFix.meta()]["0008test"], 3)
        self.assertEqual(len(result[PostFix.meta()]), 1)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import unittest

import torch
from parameterized import parameterized

from monai.networks.blocks.dints_block import ActiConvNormBlock

TEST_CASES = [
    [{"in_channel": 32, "out_channel": 16, "kernel_size": 3, "padding": 1}, (7, 32, 16, 31, 7), (7, 16, 16, 31, 7)],
    [
        {"in_channel": 32, "out_channel": 16, "kernel_size": 3, "padding": 1, "spatial_dims": 2},
        (7, 32, 13, 32),
        (7, 16, 13, 32),
    ],
]


class TestACNBlock(unittest.TestCase):

    @parameterized.expand(TEST_CASES)
    def test_acn_block(self, input_param, input_shape, expected_shape):
        net = ActiConvNormBlock(**input_param)
        result = net(torch.randn(input_shape))
        self.assertEqual(result.shape, expected_shape)


if __name__ == "__main__":
    unittest.main()



# Copyright (c) MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import unittest

import torch
from parameterized import parameterized

from monai.networks.schedulers import RFlowScheduler
from tests.test_utils import assert_allclose

TEST_2D_CASE = []
for sample_method in ["uniform", "logit-normal"]:
    TEST_2D_CASE.append(
        [{"sample_method": sample_method, "use_timestep_transform": False}, (2, 6, 16, 16), (2, 6, 16, 16)]
    )

for sample_method in ["uniform", "logit-normal"]:
    TEST_2D_CASE.append(
        [
            {"sample_method": sample_method, "use_timestep_transform": True, "spatial_dim": 2},
            (2, 6, 16, 16),
            (2, 6, 16, 16),
        ]
    )


TEST_3D_CASE = []
for sample_method in ["uniform", "logit-normal"]:
    TEST_3D_CASE.append(
        [{"sample_method": sample_method, "use_timestep_transform": False}, (2, 6, 16, 16, 16), (2, 6, 16, 16, 16)]
    )

for sample_method in ["uniform", "logit-normal"]:
    TEST_3D_CASE.append(
        [
            {"sample_method": sample_method, "use_timestep_transform": True, "spatial_dim": 3},
            (2, 6, 16, 16, 16),
            (2, 6, 16, 16, 16),
        ]
    )

TEST_CASES = TEST_2D_CASE + TEST_3D_CASE

TEST_FULl_LOOP = [
    [{"sample_method": "uniform"}, (1, 1, 2, 2), torch.Tensor([[[[-0.786166, -0.057519], [2.442662, -0.407664]]]])]
]


class TestRFlowScheduler(unittest.TestCase):
    @parameterized.expand(TEST_CASES)
    def test_add_noise(self, input_param, input_shape, expected_shape):
        scheduler = RFlowScheduler(**input_param)
        original_sample = torch.zeros(input_shape)
        timesteps = scheduler.sample_timesteps(original_sample)
        noise = torch.randn_like(original_sample)
        timesteps = torch.randint(0, scheduler.num_train_timesteps, (original_sample.shape[0],)).long()
        noisy = scheduler.add_noise(original_samples=original_sample, noise=noise, timesteps=timesteps)
        self.assertEqual(noisy.shape, expected_shape)

    @parameterized.expand(TEST_CASES)
    def test_step_shape(self, input_param, input_shape, expected_shape):
        scheduler = RFlowScheduler(**input_param)
        model_output = torch.randn(input_shape)
        sample = torch.randn(input_shape)
        scheduler.set_timesteps(num_inference_steps=100, input_img_size_numel=torch.numel(sample[0, 0, ...]))
        output_step = scheduler.step(model_output=model_output, timestep=500, sample=sample)
        self.assertEqual(output_step[0].shape, expected_shape)
        self.assertEqual(output_step[1].shape, expected_shape)

    @parameterized.expand(TEST_FULl_LOOP)
    def test_full_timestep_loop(self, input_param, input_shape, expected_output):
        scheduler = RFlowScheduler(**input_param)
        torch.manual_seed(42)
        model_output = torch.randn(input_shape)
        sample = torch.randn(input_shape)
        scheduler.set_timesteps(50, input_img_size_numel=torch.numel(sample[0, 0, ...]))
        for t in range(50):
            sample, _ = scheduler.step(model_output=model_output, timestep=t, sample=sample)
        assert_allclose(sample, expected_output, rtol=1e-3, atol=1e-3)

    def test_set_timesteps(self):
        scheduler = RFlowScheduler(num_train_timesteps=1000)
        scheduler.set_timesteps(num_inference_steps=100, input_img_size_numel=16 * 16 * 16)
        self.assertEqual(scheduler.num_inference_steps, 100)
        self.assertEqual(len(scheduler.timesteps), 100)

    def test_set_timesteps_with_num_inference_steps_bigger_than_num_train_timesteps(self):
        scheduler = RFlowScheduler(num_train_timesteps=1000)
        with self.assertRaises(ValueError):
            scheduler.set_timesteps(num_inference_steps=2000, input_img_size_numel=16 * 16 * 16)


if __name__ == "__main__":
    unittest.main()
